import importlib
from contextlib import nullcontext
import numpy as np
import functools
import logging
import ray
import time
import json
from deltacat.aws import s3u as s3_utils
import deltacat
from deltacat import logs
from deltacat.compute.compactor import (
    PyArrowWriteResult,
    RoundCompletionInfo,
)
from deltacat.compute.compactor_v2.model.merge_input import MergeInput
from deltacat.compute.compactor_v2.model.merge_result import MergeResult
from deltacat.compute.compactor_v2.model.hash_bucket_input import HashBucketInput
from deltacat.compute.compactor_v2.model.hash_bucket_result import HashBucketResult
from deltacat.compute.compactor.model.materialize_result import MaterializeResult
from deltacat.storage import (
    Delta,
    DeltaLocator,
    Partition,
)
from deltacat.compute.compactor.model.compact_partition_params import (
    CompactPartitionParams,
)
from deltacat.utils.ray_utils.concurrency import (
    invoke_parallel,
    task_resource_options_provider,
)
from deltacat.compute.compactor_v2.steps import merge as mg
from deltacat.compute.compactor_v2.steps import hash_bucket as hb
from deltacat.compute.compactor_v2.utils import io
from deltacat.compute.compactor.utils import round_completion_file as rcf

from typing import List, Optional, Tuple
from collections import defaultdict
from deltacat.compute.compactor.model.compaction_session_audit_info import (
    CompactionSessionAuditInfo,
)
from deltacat.utils.resources import (
    get_current_node_peak_memory_usage_in_bytes,
)
from deltacat.compute.compactor_v2.utils.task_options import (
    hash_bucket_resource_options_provider,
    merge_resource_options_provider,
)
from deltacat.utils.resources import ClusterUtilizationOverTimeRange
from deltacat.compute.compactor.model.compactor_version import CompactorVersion

if importlib.util.find_spec("memray"):
    import memray


logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def compact_partition(params: CompactPartitionParams, **kwargs) -> Optional[str]:

    assert (
        params.hash_bucket_count is not None and params.hash_bucket_count >= 1
    ), "hash_bucket_count is a required arg for compactor v2"

    with memray.Tracker(
        f"compaction_partition.bin"
    ) if params.enable_profiler else nullcontext(), ClusterUtilizationOverTimeRange() as cluster_util:
        (new_partition, new_rci, new_rcf_partition_locator,) = _execute_compaction(
            params,
            cluster_util=cluster_util,
            **kwargs,
        )

        logger.info(
            f"Partition-{params.source_partition_locator} -> "
            f"Compaction session data processing completed"
        )
        round_completion_file_s3_url = None
        if new_partition:
            logger.info(f"Committing compacted partition to: {new_partition.locator}")
            partition = params.deltacat_storage.commit_partition(
                new_partition, **params.deltacat_storage_kwargs
            )
            logger.info(f"Committed compacted partition: {partition}")

            round_completion_file_s3_url = rcf.write_round_completion_file(
                params.compaction_artifact_s3_bucket,
                new_rcf_partition_locator,
                new_rci,
                **params.s3_client_kwargs,
            )
        else:
            logger.warn("No new partition was committed during compaction.")

        logger.info(
            f"Completed compaction session for: {params.source_partition_locator}"
        )
        return round_completion_file_s3_url


def _execute_compaction(
    params: CompactPartitionParams, **kwargs
) -> Tuple[Optional[Partition], Optional[RoundCompletionInfo], Optional[str]]:

    rcf_source_partition_locator = (
        params.rebase_source_partition_locator or params.source_partition_locator
    )

    base_audit_url = rcf_source_partition_locator.path(
        f"s3://{params.compaction_artifact_s3_bucket}/compaction-audit"
    )
    audit_url = f"{base_audit_url}.json"
    logger.info(f"Compaction audit will be written to {audit_url}")
    compaction_audit = (
        CompactionSessionAuditInfo(deltacat.__version__, ray.__version__, audit_url)
        .set_hash_bucket_count(params.hash_bucket_count)
        .set_compactor_version(CompactorVersion.V2.value)
    )

    compaction_start = time.monotonic()

    task_max_parallelism = params.task_max_parallelism

    if params.pg_config:
        logger.info(
            "pg_config specified. Tasks will be scheduled in a placement group."
        )
        cluster_resources = params.pg_config.resource
        cluster_cpus = cluster_resources["CPU"]
        cluster_memory = cluster_resources["memory"]
        task_max_parallelism = cluster_cpus
        compaction_audit.set_total_cluster_memory_bytes(cluster_memory)

    # read the results from any previously completed compaction round
    round_completion_info = None
    high_watermark = None
    previous_compacted_delta_manifest = None

    if not params.rebase_source_partition_locator:
        round_completion_info = rcf.read_round_completion_file(
            params.compaction_artifact_s3_bucket,
            params.source_partition_locator,
            **params.s3_client_kwargs,
        )
        if not round_completion_info:
            logger.info(
                f"Both rebase partition and round completion file not found. Performing an entire backfill on source."
            )
        else:
            compacted_delta_locator = round_completion_info.compacted_delta_locator

            previous_compacted_delta_manifest = (
                params.deltacat_storage.get_delta_manifest(
                    compacted_delta_locator, **params.deltacat_storage_kwargs
                )
            )

            high_watermark = round_completion_info.high_watermark
            logger.info(f"Setting round completion high watermark: {high_watermark}")
            assert (
                params.hash_bucket_count == round_completion_info.hash_bucket_count
            ), (
                "The hash bucket count has changed. "
                "Kindly run rebase compaction and trigger incremental again. "
                f"Hash Bucket count in RCF={round_completion_info.hash_bucket_count} "
                f"not equal to Hash bucket count in args={params.hash_bucket_count}."
            )

        logger.info(f"Round completion file: {round_completion_info}")

    delta_discovery_start = time.monotonic()

    input_deltas = io.discover_deltas(
        params.source_partition_locator,
        params.last_stream_position_to_compact,
        params.rebase_source_partition_locator,
        params.rebase_source_partition_high_watermark,
        high_watermark,
        params.deltacat_storage,
        params.deltacat_storage_kwargs,
        params.list_deltas_kwargs,
    )

    uniform_deltas = io.create_uniform_input_deltas(
        input_deltas=input_deltas,
        hash_bucket_count=params.hash_bucket_count,
        compaction_audit=compaction_audit,
        deltacat_storage=params.deltacat_storage,
        previous_inflation=params.previous_inflation,
        min_delta_bytes=params.min_delta_bytes_in_batch,
        min_file_counts=params.min_files_in_batch,
        # disable input split during rebase as the rebase files are already uniform
        enable_input_split=params.rebase_source_partition_locator is None,
        deltacat_storage_kwargs=params.deltacat_storage_kwargs,
    )

    delta_discovery_end = time.monotonic()

    compaction_audit.set_uniform_deltas_created(len(uniform_deltas))
    compaction_audit.set_delta_discovery_time_in_seconds(
        delta_discovery_end - delta_discovery_start
    )

    s3_utils.upload(
        compaction_audit.audit_url,
        str(json.dumps(compaction_audit)),
        **params.s3_client_kwargs,
    )

    if not input_deltas:
        logger.info("No input deltas found to compact.")
        return None, None, None

    hb_options_provider = functools.partial(
        task_resource_options_provider,
        pg_config=params.pg_config,
        resource_amount_provider=hash_bucket_resource_options_provider,
        previous_inflation=params.previous_inflation,
        average_record_size_bytes=params.average_record_size_bytes,
        primary_keys=params.primary_keys,
        ray_custom_resources=params.ray_custom_resources,
    )

    hb_start = time.monotonic()

    def hash_bucket_input_provider(index, item):
        return {
            "input": HashBucketInput.of(
                item,
                primary_keys=params.primary_keys,
                num_hash_buckets=params.hash_bucket_count,
                num_hash_groups=params.hash_group_count,
                enable_profiler=params.enable_profiler,
                metrics_config=params.metrics_config,
                read_kwargs_provider=params.read_kwargs_provider,
                object_store=params.object_store,
                deltacat_storage=params.deltacat_storage,
                deltacat_storage_kwargs=params.deltacat_storage_kwargs,
            )
        }

    hb_tasks_pending = invoke_parallel(
        items=uniform_deltas,
        ray_task=hb.hash_bucket,
        max_parallelism=task_max_parallelism,
        options_provider=hb_options_provider,
        kwargs_provider=hash_bucket_input_provider,
    )

    hb_invoke_end = time.monotonic()

    logger.info(f"Getting {len(hb_tasks_pending)} hash bucket results...")
    hb_results: List[HashBucketResult] = ray.get(hb_tasks_pending)
    logger.info(f"Got {len(hb_results)} hash bucket results.")
    hb_end = time.monotonic()

    # we use time.time() here because time.monotonic() has no reference point
    # whereas time.time() measures epoch seconds. Hence, it will be reasonable
    # to compare time.time()s captured in different nodes.
    hb_results_retrieved_at = time.time()

    telemetry_time_hb = compaction_audit.save_step_stats(
        CompactionSessionAuditInfo.HASH_BUCKET_STEP_NAME,
        hb_results,
        hb_results_retrieved_at,
        hb_invoke_end - hb_start,
        hb_end - hb_start,
    )

    s3_utils.upload(
        compaction_audit.audit_url,
        str(json.dumps(compaction_audit)),
        **params.s3_client_kwargs,
    )

    all_hash_group_idx_to_obj_id = defaultdict(list)
    all_hash_group_idx_to_size_bytes = defaultdict(int)
    all_hash_group_idx_to_num_rows = defaultdict(int)
    hb_data_processed_size_bytes = np.int64(0)
    total_hb_record_count = np.int64(0)

    # initialize all hash groups
    for hb_group in range(params.hash_group_count):
        all_hash_group_idx_to_num_rows[hb_group] = 0
        all_hash_group_idx_to_obj_id[hb_group] = []
        all_hash_group_idx_to_size_bytes[hb_group] = 0

    for hb_result in hb_results:
        hb_data_processed_size_bytes += hb_result.hb_size_bytes
        total_hb_record_count += hb_result.hb_record_count

        for hash_group_index, object_id_size_tuple in enumerate(
            hb_result.hash_bucket_group_to_obj_id_tuple
        ):
            if object_id_size_tuple:
                all_hash_group_idx_to_obj_id[hash_group_index].append(
                    object_id_size_tuple[0]
                )
                all_hash_group_idx_to_size_bytes[
                    hash_group_index
                ] += object_id_size_tuple[1].item()
                all_hash_group_idx_to_num_rows[
                    hash_group_index
                ] += object_id_size_tuple[2].item()

    logger.info(
        f"Got {total_hb_record_count} hash bucket records from hash bucketing step..."
    )

    compaction_audit.set_input_records(total_hb_record_count.item())
    compaction_audit.set_hash_bucket_processed_size_bytes(
        hb_data_processed_size_bytes.item()
    )

    # create a new stream for this round
    compacted_stream_locator = params.destination_partition_locator.stream_locator
    compacted_stream = params.deltacat_storage.get_stream(
        compacted_stream_locator.namespace,
        compacted_stream_locator.table_name,
        compacted_stream_locator.table_version,
        **params.deltacat_storage_kwargs,
    )
    compacted_partition = params.deltacat_storage.stage_partition(
        compacted_stream,
        params.destination_partition_locator.partition_values,
        **params.deltacat_storage_kwargs,
    )

    # BSP Step 2: Merge
    merge_options_provider = functools.partial(
        task_resource_options_provider,
        pg_config=params.pg_config,
        resource_amount_provider=merge_resource_options_provider,
        num_hash_groups=params.hash_group_count,
        hash_group_size_bytes=all_hash_group_idx_to_size_bytes,
        hash_group_num_rows=all_hash_group_idx_to_num_rows,
        round_completion_info=round_completion_info,
        compacted_delta_manifest=previous_compacted_delta_manifest,
        primary_keys=params.primary_keys,
        deltacat_storage=params.deltacat_storage,
        deltacat_storage_kwargs=params.deltacat_storage_kwargs,
        ray_custom_resources=params.ray_custom_resources,
    )

    def merge_input_provider(index, item):
        return {
            "input": MergeInput.of(
                dfe_groups_refs=item[1],
                write_to_partition=compacted_partition,
                compacted_file_content_type=params.compacted_file_content_type,
                primary_keys=params.primary_keys,
                sort_keys=params.sort_keys,
                merge_task_index=index,
                hash_bucket_count=params.hash_bucket_count,
                drop_duplicates=params.drop_duplicates,
                hash_group_index=item[0],
                num_hash_groups=params.hash_group_count,
                max_records_per_output_file=params.records_per_compacted_file,
                enable_profiler=params.enable_profiler,
                metrics_config=params.metrics_config,
                s3_table_writer_kwargs=params.s3_table_writer_kwargs,
                read_kwargs_provider=params.read_kwargs_provider,
                round_completion_info=round_completion_info,
                object_store=params.object_store,
                deltacat_storage=params.deltacat_storage,
                deltacat_storage_kwargs=params.deltacat_storage_kwargs,
            )
        }

    merge_start = time.monotonic()

    merge_tasks_pending = invoke_parallel(
        items=all_hash_group_idx_to_obj_id.items(),
        ray_task=mg.merge,
        max_parallelism=task_max_parallelism,
        options_provider=merge_options_provider,
        kwargs_provider=merge_input_provider,
    )

    merge_invoke_end = time.monotonic()
    logger.info(f"Getting {len(merge_tasks_pending)} merge results...")
    merge_results: List[MergeResult] = ray.get(merge_tasks_pending)
    logger.info(f"Got {len(merge_results)} merge results.")

    merge_results_retrieved_at = time.time()
    merge_end = time.monotonic()

    total_dd_record_count = sum([ddr.deduped_record_count for ddr in merge_results])
    logger.info(f"Deduped {total_dd_record_count} records...")

    telemetry_time_merge = compaction_audit.save_step_stats(
        CompactionSessionAuditInfo.MERGE_STEP_NAME,
        merge_results,
        merge_results_retrieved_at,
        merge_invoke_end - merge_start,
        merge_end - merge_start,
    )

    compaction_audit.set_records_deduped(total_dd_record_count.item())

    mat_results = []
    for merge_result in merge_results:
        mat_results.extend(merge_result.materialize_results)

    mat_results: List[MaterializeResult] = sorted(
        mat_results, key=lambda m: m.task_index
    )

    hb_id_to_entry_indices_range = {}
    file_index = 0
    previous_task_index = -1

    for mat_result in mat_results:
        assert (
            mat_result.pyarrow_write_result.files >= 1
        ), "Atleast one file must be materialized"
        assert (
            mat_result.task_index != previous_task_index
        ), f"Multiple materialize results found for a hash bucket: {mat_result.task_index}"

        hb_id_to_entry_indices_range[str(mat_result.task_index)] = (
            file_index,
            file_index + mat_result.pyarrow_write_result.files,
        )

        file_index += mat_result.pyarrow_write_result.files
        previous_task_index = mat_result.task_index

    s3_utils.upload(
        compaction_audit.audit_url,
        str(json.dumps(compaction_audit)),
        **params.s3_client_kwargs,
    )

    deltas = [m.delta for m in mat_results]

    # Note: An appropriate last stream position must be set
    # to avoid correctness issue.
    merged_delta = Delta.merge_deltas(
        deltas,
        stream_position=params.last_stream_position_to_compact,
    )

    record_info_msg = (
        f"Hash bucket records: {total_hb_record_count},"
        f" Deduped records: {total_dd_record_count}, "
        f" Materialized records: {merged_delta.meta.record_count}"
    )
    logger.info(record_info_msg)

    compacted_delta = params.deltacat_storage.commit_delta(
        merged_delta,
        properties=kwargs.get("properties", {}),
        **params.deltacat_storage_kwargs,
    )

    logger.info(f"Committed compacted delta: {compacted_delta}")

    compaction_end = time.monotonic()
    compaction_audit.set_compaction_time_in_seconds(compaction_end - compaction_start)

    new_compacted_delta_locator = DeltaLocator.of(
        compacted_partition.locator,
        compacted_delta.stream_position,
    )

    pyarrow_write_result = PyArrowWriteResult.union(
        [m.pyarrow_write_result for m in mat_results]
    )

    session_peak_memory = get_current_node_peak_memory_usage_in_bytes()
    compaction_audit.set_peak_memory_used_bytes_by_compaction_session_process(
        session_peak_memory
    )

    compaction_audit.save_round_completion_stats(
        mat_results, telemetry_time_hb + telemetry_time_merge
    )

    cluster_util: ClusterUtilizationOverTimeRange = kwargs.get("cluster_util")

    if cluster_util:
        compaction_audit.set_total_cpu_seconds(cluster_util.total_vcpu_seconds)
        compaction_audit.set_used_cpu_seconds(cluster_util.used_vcpu_seconds)
        compaction_audit.set_used_memory_gb_seconds(cluster_util.used_memory_gb_seconds)
        compaction_audit.set_total_memory_gb_seconds(
            cluster_util.total_memory_gb_seconds
        )
        compaction_audit.set_cluster_cpu_max(cluster_util.max_cpu)

    input_inflation = None
    input_average_record_size_bytes = None
    # Note: we only consider inflation for incremental delta
    if (
        compaction_audit.input_size_bytes
        and compaction_audit.hash_bucket_processed_size_bytes
    ):
        input_inflation = (
            compaction_audit.hash_bucket_processed_size_bytes
            / compaction_audit.input_size_bytes
        )

    if (
        compaction_audit.hash_bucket_processed_size_bytes
        and compaction_audit.input_records
    ):
        input_average_record_size_bytes = (
            compaction_audit.hash_bucket_processed_size_bytes
            / compaction_audit.input_records
        )

    logger.info(
        f"The inflation of input deltas={input_inflation}"
        f" and average record size={input_average_record_size_bytes}"
    )

    # After all incremental delta related calculations, we update
    # the input sizes to accomodate the compacted table
    if round_completion_info:
        compaction_audit.set_input_file_count(
            (compaction_audit.input_file_count or 0)
            + round_completion_info.compacted_pyarrow_write_result.files
        )
        compaction_audit.set_input_size_bytes(
            (compaction_audit.input_size_bytes or 0.0)
            + round_completion_info.compacted_pyarrow_write_result.file_bytes
        )
        compaction_audit.set_input_records(
            (compaction_audit.input_records or 0)
            + round_completion_info.compacted_pyarrow_write_result.records
        )

    s3_utils.upload(
        compaction_audit.audit_url,
        str(json.dumps(compaction_audit)),
        **params.s3_client_kwargs,
    )

    new_round_completion_info = RoundCompletionInfo.of(
        high_watermark=params.last_stream_position_to_compact,
        compacted_delta_locator=new_compacted_delta_locator,
        compacted_pyarrow_write_result=pyarrow_write_result,
        sort_keys_bit_width=params.bit_width_of_sort_keys,
        manifest_entry_copied_by_reference_ratio=compaction_audit.untouched_file_ratio,
        compaction_audit_url=audit_url,
        hash_bucket_count=params.hash_bucket_count,
        hb_index_to_entry_range=hb_id_to_entry_indices_range,
        compactor_version=CompactorVersion.V2.value,
        input_inflation=input_inflation,
        input_average_record_size_bytes=input_average_record_size_bytes,
    )

    logger.info(
        f"partition-{params.source_partition_locator.partition_values},"
        f"compacted at: {params.last_stream_position_to_compact},"
    )

    return (
        compacted_partition,
        new_round_completion_info,
        rcf_source_partition_locator,
    )



import logging
from ray import cloudpickle
import time
from deltacat.io.object_store import IObjectStore
from typing import Any, List
from deltacat import logs
import uuid
from deltacat.aws import s3u as s3_utils

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class S3ObjectStore(IObjectStore):
    """
    An implementation of object store that uses S3.
    """

    def __init__(self, bucket_prefix: str) -> None:
        self.bucket = bucket_prefix
        super().__init__()

    def put_many(self, objects: List[object], *args, **kwargs) -> List[Any]:
        result = []
        for obj in objects:
            serialized = cloudpickle.dumps(obj)
            ref = uuid.uuid4()

            s3_utils.upload(f"s3://{self.bucket}/{ref}", serialized)
            result.append(ref)

        return result

    def get_many(self, refs: List[Any], *args, **kwargs) -> List[object]:
        result = []
        start = time.monotonic()
        for ref in refs:
            cur = s3_utils.download(f"s3://{self.bucket}/{ref}")
            serialized = cur["Body"].read()
            loaded = cloudpickle.loads(serialized)
            result.append(loaded)
        end = time.monotonic()

        logger.info(f"The total time taken to read all objects is: {end - start}")
        return result



from typing import Any, Callable, Dict, List, Optional, Union

import pyarrow as pa
import s3fs
from ray.data import read_datasource
from ray.data._internal.arrow_block import ArrowRow

from deltacat import ContentType
from deltacat.io.aws.redshift.redshift_datasource import (
    HivePartitionParser,
    RedshiftDatasource,
    RedshiftUnloadTextArgs,
    S3PathType,
)
from deltacat.io.dataset import DeltacatDataset
from deltacat.utils.common import ReadKwargsProvider


def read_redshift(
    paths: Union[str, List[str]],
    *,
    path_type: S3PathType = S3PathType.MANIFEST,
    filesystem: Optional[Union[pa.fs.S3FileSystem, s3fs.S3FileSystem]] = None,
    columns: Optional[List[str]] = None,
    schema: Optional[pa.Schema] = None,
    unload_text_args: RedshiftUnloadTextArgs = RedshiftUnloadTextArgs(),
    partitioning: HivePartitionParser = None,
    content_type_provider: Callable[[str], ContentType] = lambda p: ContentType.PARQUET
    if p.endswith(".parquet")
    else ContentType.CSV,
    parallelism: int = 200,
    ray_remote_args: Dict[str, Any] = None,
    arrow_open_stream_args: Optional[Dict[str, Any]] = None,
    pa_read_func_kwargs_provider: Optional[ReadKwargsProvider] = None,
    **kwargs,
) -> DeltacatDataset[ArrowRow]:
    """Reads Redshift UNLOAD results from either S3 Parquet or delimited text
    files into a Ray Dataset.

    Examples:
        >>> # Read all files contained in a Redshift Manifest:
        >>> import deltacat as dc
        >>> dc.io.read_redshift("/bucket/dir/manifest")

        >>> # Read all files matching the given key prefix. If this prefix
        >>> # refers to multiple files, like s3://bucket/data.parquet,
        >>> # s3://bucket/data.1.csv, etc. then all will be read. The dataset
        >>> # schema will be inferred from the first parquet file and used for
        >>> # explicit type conversion of all CSV files:
        >>> dc.io.read_redshift(
        >>>     "s3://bucket/data.txt",
        >>>     path_type=S3PathType.PREFIX)

        >>> # Read all files matching the given key prefix. If this prefix
        >>> # refers to multiple files or folders, like s3://bucket/dir/,
        >>> # s3://bucket/dir1/, s3://bucket/dir.txt, s3://bucket/dir.txt.1,
        >>> # then all files and subfolder contents will be read.
        >>> dc.io.read_redshift(
        >>>     "/bucket/dir",
        >>>     path_type=S3PathType.PREFIX)

        >>> # Read multiple files and folders:
        >>> dc.io.read_redshift(
        >>>     ["/bucket/file1", "/bucket/folder1/"],
        >>>     path_type=S3PathType.FILES_AND_FOLDERS)

        >>> # Read multiple Parquet and CSV files. The dataset schema will be
        >>> # inferred from the first parquet file and used for explicit type
        >>> # conversion of all CSV files:
        >>> dc.io.read_redshift(
        >>>     ["/bucket/file.parquet", "/bucket/file.csv"],
        >>>     path_type=S3PathType.FILES_AND_FOLDERS)

    Args:
        paths: Paths to S3 files and folders to read. If `path_type` is
            `MANIFEST` then this must be an S3 Redshift Manifest JSON file. If
            `path_type` is `PREFIX` then this must be a valid S3 key prefix.
            All files matching the key prefix, including files in matching
            subdirectories, will be read. Unless custom
            `content_type_extensions` are specified, file content types will be
            inferred by file extension with ".parquet" used for Parquet files,
            and all others assumed to be delimited text (e.g. CSV). It's
            recommended to specify the path to a manifest unloaded with the
            VERBOSE option whenever possible to improve the correctness and
            performance of Dataset reads, compute operations, and writes.
            `FILES_AND_FOLDERS` is not recommended when reading thousands of
            files due to its relatively high-latency.
        path_type: Determines how the `paths` parameter is interpreted.
        filesystem: The filesystem implementation to read from. This should be
            either PyArrow's S3FileSystem or s3fs.
        columns: A list of column names to read. Reads all columns if None or
            empty.
        schema: PyArrow schema used to determine delimited text column
            names and types. If not specified and both Parquet and delimited
            text files are read as input, then the first Parquet file schema
            discovered is used instead.
        unload_text_args: Arguments used when running Redshift `UNLOAD` to
            text file formats (e.g. CSV). These arguments ensure that all input
            text files will be correctly parsed. If not specified, then all
            text files read are assumed to use Redshift UNLOAD's default
            pipe-delimited text format.
        partition_base_dir: Base directory to start searching for partitions
            (exclusive). File paths outside of this directory will not be parsed
            for partitions and automatically added to the dataset without passing
            through any partition filter. Specify `None` or an empty string to
            search for partitions in all file path directories.
        partition_filter_fn: Callback used to filter `PARTITION` columns. Receives a
            dictionary mapping partition keys to values as input, returns `True` to
            read a partition, and `False` to skip it. Each partition key and value
            is a string parsed directly from an S3 key using hive-style
            partition directory names of the form "{key}={value}". For example:
            ``lambda x:
            True if x["month"] == "January" and x["year"] == "2022" else False``
        content_type_provider: Takes a file path as input and returns the file
            content type as output.
        parallelism: The requested parallelism of the read. Parallelism may be
            limited by the number of files of the dataset.
        ray_remote_args: kwargs passed to `ray.remote` in the read tasks.
        arrow_open_stream_args: kwargs passed to to
            `pa.fs.open_input_stream()`.
        pa_read_func_kwargs_provider: Callback that takes a `ContentType` value
            string as input, and provides read options to pass to either
            `pa.csv.open_csv()` or `pa.parquet.read_table()` as output.
    Returns:
        Dataset holding Arrow records read from the specified paths.
    """
    dataset = read_datasource(
        RedshiftDatasource(),
        parallelism=parallelism,
        paths=paths,
        content_type_provider=content_type_provider,
        path_type=path_type,
        filesystem=filesystem,
        columns=columns,
        schema=schema,
        unload_args=unload_text_args,
        partitioning=partitioning,
        ray_remote_args=ray_remote_args,
        open_stream_args=arrow_open_stream_args,
        read_kwargs_provider=pa_read_func_kwargs_provider,
        **kwargs,
    )
    return DeltacatDataset.from_dataset(dataset)



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Any, Dict, List, Optional

from deltacat.aws.redshift import Manifest, ManifestAuthor, ManifestMeta
from deltacat.storage.model.locator import Locator
from deltacat.storage.model.namespace import NamespaceLocator
from deltacat.storage.model.partition import PartitionLocator
from deltacat.storage.model.stream import StreamLocator
from deltacat.storage.model.table import TableLocator
from deltacat.storage.model.table_version import TableVersionLocator
from deltacat.storage.model.types import DeltaType


class Delta(dict):
    @staticmethod
    def of(
        locator: Optional[DeltaLocator],
        delta_type: Optional[DeltaType],
        meta: Optional[ManifestMeta],
        properties: Optional[Dict[str, str]],
        manifest: Optional[Manifest],
        previous_stream_position: Optional[int] = None,
    ) -> Delta:
        """
        Creates a Delta metadata model with the given Delta Locator, Delta Type,
        manifest metadata, properties, manifest, and previous delta stream
        position.
        """
        delta = Delta()
        delta.locator = locator
        delta.type = delta_type
        delta.meta = meta
        delta.properties = properties
        delta.manifest = manifest
        delta.previous_stream_position = previous_stream_position
        return delta

    @staticmethod
    def merge_deltas(
        deltas: List[Delta],
        manifest_author: Optional[ManifestAuthor] = None,
        stream_position: Optional[int] = None,
        properties: Optional[Dict[str, str]] = None,
    ) -> Delta:
        """
        Merges the input list of deltas into a single delta. All input deltas to
        merge must belong to the same partition, share the same delta type, and
        have non-empty manifests.

        Manifest content type and content encoding will be set to None in the
        event of conflicting types or encodings between individual manifest
        entries. Missing record counts, content lengths, and source content
        lengths will be coalesced to 0. Delta properties, stream position, and
        manifest author will be set to None unless explicitly specified.

        The previous partition stream position of the merged delta is set to the
        maximum previous partition stream position of all input deltas, or None
        if the previous partition stream position of any input delta is None.

        Input delta manifest entry order will be preserved in the merged delta
        returned. That is, if manifest entry A preceded manifest entry B
        in the input delta list, then manifest entry A will precede manifest
        entry B in the merged delta.
        """
        if not deltas:
            raise ValueError("No deltas given to merge.")
        manifests = [d.manifest for d in deltas]
        if any(not m for m in manifests):
            raise ValueError(f"Deltas to merge must have non-empty manifests.")
        distinct_storage_types = set([d.storage_type for d in deltas])
        if len(distinct_storage_types) > 1:
            raise NotImplementedError(
                f"Deltas to merge must all share the same storage type "
                f"(found {len(distinct_storage_types)} storage types."
            )
        pl_digest_set = set([d.partition_locator.digest() for d in deltas])
        if len(pl_digest_set) > 1:
            raise ValueError(
                f"Deltas to merge must all belong to the same partition "
                f"(found {len(pl_digest_set)} partitions)."
            )
        distinct_delta_types = set([d.type for d in deltas])
        if len(distinct_delta_types) > 1:
            raise ValueError(
                f"Deltas to merge must all share the same delta type "
                f"(found {len(distinct_delta_types)} delta types)."
            )
        merged_manifest = Manifest.merge_manifests(
            manifests,
            manifest_author,
        )
        partition_locator = deltas[0].partition_locator
        prev_positions = [d.previous_stream_position for d in deltas]
        prev_position = None if None in prev_positions else max(prev_positions)
        return Delta.of(
            DeltaLocator.of(partition_locator, stream_position),
            distinct_delta_types.pop(),
            merged_manifest.meta,
            properties,
            merged_manifest,
            prev_position,
        )

    @property
    def manifest(self) -> Optional[Manifest]:
        val: Dict[str, Any] = self.get("manifest")
        if val is not None and not isinstance(val, Manifest):
            self.manifest = val = Manifest(val)
        return val

    @manifest.setter
    def manifest(self, manifest: Optional[Manifest]) -> None:
        self["manifest"] = manifest

    @property
    def meta(self) -> Optional[ManifestMeta]:
        val: Dict[str, Any] = self.get("meta")
        if val is not None and not isinstance(val, ManifestMeta):
            self.meta = val = ManifestMeta(val)
        return val

    @meta.setter
    def meta(self, meta: Optional[ManifestMeta]) -> None:
        self["meta"] = meta

    @property
    def properties(self) -> Optional[Dict[str, str]]:
        return self.get("properties")

    @properties.setter
    def properties(self, properties: Optional[Dict[str, str]]) -> None:
        self["properties"] = properties

    @property
    def type(self) -> Optional[DeltaType]:
        delta_type = self.get("type")
        return None if delta_type is None else DeltaType(delta_type)

    @type.setter
    def type(self, delta_type: Optional[DeltaType]) -> None:
        self["type"] = delta_type

    @property
    def locator(self) -> Optional[DeltaLocator]:
        val: Dict[str, Any] = self.get("deltaLocator")
        if val is not None and not isinstance(val, DeltaLocator):
            self.locator = val = DeltaLocator(val)
        return val

    @locator.setter
    def locator(self, delta_locator: Optional[DeltaLocator]) -> None:
        self["deltaLocator"] = delta_locator

    @property
    def previous_stream_position(self) -> Optional[int]:
        return self.get("previousStreamPosition")

    @previous_stream_position.setter
    def previous_stream_position(self, previous_stream_position: Optional[int]) -> None:
        self["previousStreamPosition"] = previous_stream_position

    @property
    def namespace_locator(self) -> Optional[NamespaceLocator]:
        delta_locator = self.locator
        if delta_locator:
            return delta_locator.namespace_locator
        return None

    @property
    def table_locator(self) -> Optional[TableLocator]:
        delta_locator = self.locator
        if delta_locator:
            return delta_locator.table_locator
        return None

    @property
    def table_version_locator(self) -> Optional[TableVersionLocator]:
        delta_locator = self.locator
        if delta_locator:
            return delta_locator.table_version_locator
        return None

    @property
    def stream_locator(self) -> Optional[StreamLocator]:
        delta_locator = self.locator
        if delta_locator:
            return delta_locator.stream_locator
        return None

    @property
    def partition_locator(self) -> Optional[PartitionLocator]:
        delta_locator = self.locator
        if delta_locator:
            return delta_locator.partition_locator
        return None

    @property
    def storage_type(self) -> Optional[str]:
        delta_locator = self.locator
        if delta_locator:
            return delta_locator.storage_type
        return None

    @property
    def namespace(self) -> Optional[str]:
        delta_locator = self.locator
        if delta_locator:
            return delta_locator.namespace
        return None

    @property
    def table_name(self) -> Optional[str]:
        delta_locator = self.locator
        if delta_locator:
            return delta_locator.table_name
        return None

    @property
    def table_version(self) -> Optional[str]:
        delta_locator = self.locator
        if delta_locator:
            return delta_locator.table_version
        return None

    @property
    def stream_id(self) -> Optional[str]:
        delta_locator = self.locator
        if delta_locator:
            return delta_locator.stream_id
        return None

    @property
    def partition_id(self) -> Optional[str]:
        delta_locator = self.locator
        if delta_locator:
            return delta_locator.partition_id
        return None

    @property
    def partition_values(self) -> Optional[List[Any]]:
        delta_locator = self.locator
        if delta_locator:
            return delta_locator.partition_values
        return None

    @property
    def stream_position(self) -> Optional[int]:
        delta_locator = self.locator
        if delta_locator:
            return delta_locator.stream_position
        return None


class DeltaLocator(Locator, dict):
    @staticmethod
    def of(
        partition_locator: Optional[PartitionLocator] = None,
        stream_position: Optional[int] = None,
    ) -> DeltaLocator:
        """
        Creates a partition delta locator. Stream Position, if provided, should
        be greater than that of any prior delta in the partition.
        """
        delta_locator = DeltaLocator()
        delta_locator.partition_locator = partition_locator
        delta_locator.stream_position = stream_position
        return delta_locator

    @staticmethod
    def at(
        namespace: Optional[str],
        table_name: Optional[str],
        table_version: Optional[str],
        stream_id: Optional[str],
        storage_type: Optional[str],
        partition_values: Optional[List[Any]],
        partition_id: Optional[str],
        stream_position: Optional[int],
    ) -> DeltaLocator:
        partition_locator = PartitionLocator.at(
            namespace,
            table_name,
            table_version,
            stream_id,
            storage_type,
            partition_values,
            partition_id,
        )
        return DeltaLocator.of(
            partition_locator,
            stream_position,
        )

    @property
    def partition_locator(self) -> Optional[PartitionLocator]:
        val: Dict[str, Any] = self.get("partitionLocator")
        if val is not None and not isinstance(val, PartitionLocator):
            self.partition_locator = val = PartitionLocator(val)
        return val

    @partition_locator.setter
    def partition_locator(self, partition_locator: Optional[PartitionLocator]) -> None:
        self["partitionLocator"] = partition_locator

    @property
    def stream_position(self) -> Optional[int]:
        return self.get("streamPosition")

    @stream_position.setter
    def stream_position(self, stream_position: Optional[int]) -> None:
        self["streamPosition"] = stream_position

    @property
    def namespace_locator(self) -> Optional[NamespaceLocator]:
        partition_locator = self.partition_locator
        if partition_locator:
            return partition_locator.namespace_locator
        return None

    @property
    def table_locator(self) -> Optional[TableLocator]:
        partition_locator = self.partition_locator
        if partition_locator:
            return partition_locator.table_locator
        return None

    @property
    def table_version_locator(self) -> Optional[TableVersionLocator]:
        partition_locator = self.partition_locator
        if partition_locator:
            return partition_locator.table_version_locator
        return None

    @property
    def stream_locator(self) -> Optional[StreamLocator]:
        partition_locator = self.partition_locator
        if partition_locator:
            return partition_locator.stream_locator
        return None

    @property
    def partition_values(self) -> Optional[List[Any]]:
        partition_locator = self.partition_locator
        if partition_locator:
            return partition_locator.partition_values
        return None

    @property
    def partition_id(self) -> Optional[str]:
        partition_locator = self.partition_locator
        if partition_locator:
            return partition_locator.partition_id
        return None

    @property
    def stream_id(self) -> Optional[str]:
        partition_locator = self.partition_locator
        if partition_locator:
            return partition_locator.stream_id
        return None

    @property
    def storage_type(self) -> Optional[str]:
        partition_locator = self.partition_locator
        if partition_locator:
            return partition_locator.storage_type
        return None

    @property
    def namespace(self) -> Optional[str]:
        partition_locator = self.partition_locator
        if partition_locator:
            return partition_locator.namespace
        return None

    @property
    def table_name(self) -> Optional[str]:
        partition_locator = self.partition_locator
        if partition_locator:
            return partition_locator.table_name
        return None

    @property
    def table_version(self) -> Optional[str]:
        partition_locator = self.partition_locator
        if partition_locator:
            return partition_locator.table_version
        return None

    def canonical_string(self) -> str:
        """
        Returns a unique string for the given locator that can be used
        for equality checks (i.e. two locators are equal if they have
        the same canonical string).
        """
        pl_hexdigest = self.partition_locator.hexdigest()
        stream_position = self.stream_position
        return f"{pl_hexdigest}|{stream_position}"



import logging
from functools import lru_cache
from typing import Optional
from http import HTTPStatus

import boto3
from boto3.exceptions import ResourceNotExistsError
from boto3.resources.base import ServiceResource
from botocore.client import BaseClient
from botocore.config import Config
from requests.adapters import Response
from tenacity import (
    RetryError,
    Retrying,
    wait_fixed,
    retry_if_exception,
    stop_after_delay,
)

from deltacat import logs
from deltacat.aws.constants import BOTO_MAX_RETRIES
import requests


logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

BOTO3_PROFILE_NAME_KWARG_KEY = "boto3_profile_name"
INSTANCE_METADATA_SERVICE_IPV4_URI = "http://169.254.169.254/latest/meta-data/"  # https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/instancedata-data-retrieval.html
RETRYABLE_HTTP_STATUS_CODES = [
    # 429
    HTTPStatus.TOO_MANY_REQUESTS,
    # 5xx
    HTTPStatus.INTERNAL_SERVER_ERROR,
    HTTPStatus.NOT_IMPLEMENTED,
    HTTPStatus.BAD_GATEWAY,
    HTTPStatus.SERVICE_UNAVAILABLE,
    HTTPStatus.GATEWAY_TIMEOUT,
]


class RetryIfRetryableHTTPStatusCode(retry_if_exception):
    """
    Retry strategy that retries if the exception is an ``HTTPError`` with
    a status code in the retryable errors list.
    """

    def __init__(self):
        def is_retryable_error(exception):
            return isinstance(exception, requests.exceptions.ConnectionError) or (
                isinstance(exception, requests.exceptions.HTTPError)
                and exception.response.status_code in RETRYABLE_HTTP_STATUS_CODES
            )

        super().__init__(predicate=is_retryable_error)


def _log_attempt_number(retry_state):
    """return the result of the last call attempt"""
    logger.warning(f"Retrying: {retry_state.attempt_number}...")


def _get_url(url: str, get_url_kwargs=None):
    if get_url_kwargs is None:
        get_url_kwargs = {}
    resp = requests.get(url, **get_url_kwargs)
    resp.raise_for_status()
    return resp


def retrying_get(
    url: str,
    retry_strategy,
    wait_strategy,
    stop_strategy,
) -> Optional[Response]:
    """Retries a request to the given URL until it succeeds.

    Args:
        retry_strategy (Callable): A function that returns a retry strategy.
        wait_strategy (Callable): A function that returns a wait strategy.
        stop_strategy (Callable): A function that returns a stop strategy.
        url (str): The URL to retry.

    Returns:
        Optional[Response]: The response from the URL, or None if the request
            failed after the maximum number of retries.
    """
    try:
        for attempt in Retrying(
            retry=retry_strategy(),
            wait=wait_strategy,
            stop=stop_strategy,
            after=_log_attempt_number,
        ):
            with attempt:
                resp = _get_url(url)
                return resp
    except RetryError as re:
        logger.error(f"Failed to retry URL: {url} - {re}")
    logger.info(f"Unable to get from URL: {url}")
    return None


def block_until_instance_metadata_service_returns_success(
    url=INSTANCE_METADATA_SERVICE_IPV4_URI,
    retry_strategy=RetryIfRetryableHTTPStatusCode,
    wait_strategy=wait_fixed(2),  # wait 2 seconds before retrying,
    stop_strategy=stop_after_delay(60 * 30),  # stop trying after 30 minutes
) -> Optional[Response]:
    """Blocks until the instance metadata service returns a successful response.

    Args:
        retry_strategy (Callable): A function that returns a retry strategy.
        wait_strategy (Callable): A function that returns a wait strategy.
        stop_strategy (Callable): A function that returns a stop strategy.
        url (str): The URL of the instance metadata service.

    Returns:
        Optional[Response]: The response from the instance metadata service,
            or None if the request failed after the maximum number of retries.

    https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/instancedata-data-retrieval.html
    """
    # We will get a 403 HTTP status code if running deltacat not in an EC2 instance. In that case we won't want to block.
    return retrying_get(
        url,
        retry_strategy,
        wait_strategy,
        stop_strategy,
    )


def _get_session_from_kwargs(input_kwargs):
    block_until_instance_metadata_service_returns_success()
    if input_kwargs.get(BOTO3_PROFILE_NAME_KWARG_KEY) is not None:
        boto3_session = boto3.Session(
            profile_name=input_kwargs.get(BOTO3_PROFILE_NAME_KWARG_KEY)
        )
        input_kwargs.pop(BOTO3_PROFILE_NAME_KWARG_KEY)
        return boto3_session
    else:
        return boto3.Session()


def _resource(name: str, region: Optional[str], **kwargs) -> ServiceResource:
    boto3_session = _get_session_from_kwargs(kwargs)

    boto_config = Config(retries={"max_attempts": BOTO_MAX_RETRIES, "mode": "adaptive"})
    kwargs = {"config": boto_config, **kwargs}
    return boto3_session.resource(
        name,
        region,
        **kwargs,
    )


def _client(name: str, region: Optional[str], **kwargs) -> BaseClient:
    try:
        # try to re-use a client from the resource cache first
        return resource_cache(name, region, **kwargs).meta.client
    except ResourceNotExistsError:
        # fall back for clients without an associated resource
        boto3_session = _get_session_from_kwargs(kwargs)
        boto_config = Config(
            retries={"max_attempts": BOTO_MAX_RETRIES, "mode": "adaptive"}
        )
        kwargs = {"config": boto_config, **kwargs}
        return boto3_session.client(
            name,
            region,
            **kwargs,
        )


def resource_cache(name: str, region: Optional[str], **kwargs) -> ServiceResource:
    # we don't use the @lru_cache decorator because Ray can't pickle it
    cached_function = lru_cache()(_resource)
    return cached_function(name, region, **kwargs)


def client_cache(name: str, region: Optional[str], **kwargs) -> BaseClient:
    # we don't use the @lru_cache decorator because Ray can't pickle it
    cached_function = lru_cache()(_client)
    return cached_function(name, region, **kwargs)



import logging
from deltacat import logs
import pyarrow.compute as pc
import pyarrow as pa
import numpy as np
from deltacat.utils.performance import timed_invocation
from deltacat.compute.compactor.utils import system_columns as sc

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def _create_chunked_index_array(array: pa.Array) -> pa.Array:
    """
    Creates an chunked array where each chunk is of same size in the input array.
    """
    chunk_lengths = [
        len(array.chunk(chunk_index)) for chunk_index in range(len(array.chunks))
    ]

    # if all chunk lengths are equal, numpy assumes correct shape
    # which make addition operation append elements instead of adding.
    result = np.empty((len(chunk_lengths),), dtype="object")

    for index, cl in enumerate(chunk_lengths):
        result[index] = np.arange(cl, dtype="int32")

    chunk_lengths = ([0] + chunk_lengths)[:-1]
    result = pa.chunked_array(result + np.cumsum(chunk_lengths))
    return result


def drop_duplicates(table: pa.Table, on: str) -> pa.Table:
    """
    It is important to not combine the chunks for performance reasons.
    """

    if on not in table.column_names:
        return table

    index_array, array_latency = timed_invocation(
        _create_chunked_index_array, table[on]
    )

    logger.info(
        "Created a chunked index array of length "
        f" {len(index_array)} in {array_latency}s"
    )

    table = table.set_column(
        table.shape[1], sc._ORDERED_RECORD_IDX_COLUMN_NAME, index_array
    )
    selector = table.group_by([on]).aggregate(
        [(sc._ORDERED_RECORD_IDX_COLUMN_NAME, "max")]
    )

    table = table.filter(
        pc.is_in(
            table[sc._ORDERED_RECORD_IDX_COLUMN_NAME],
            value_set=selector[f"{sc._ORDERED_RECORD_IDX_COLUMN_NAME}_max"],
        )
    )

    table = table.drop([sc._ORDERED_RECORD_IDX_COLUMN_NAME])

    return table



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Tuple
from deltacat.storage import DeltaLocator, PartitionLocator
from deltacat.compute.compactor.model.pyarrow_write_result import PyArrowWriteResult
from typing import Any, Dict, Optional


class HighWatermark(dict):
    """
    Inherit from dict to make it easy for serialization/deserialization.
    Keep both partition locator and high watermark as a tuple to be persisted in the rcf
    """

    def set(self, partition_locator: PartitionLocator, delta_stream_position: int):
        self[partition_locator.canonical_string()] = (
            partition_locator,
            delta_stream_position,
        )

    def get(self, partition_locator: PartitionLocator) -> int:
        if partition_locator.canonical_string() in self:
            return self[partition_locator.canonical_string()][1]
        return 0


class RoundCompletionInfo(dict):
    """
    In case of compacting deltas from multi-sources, the high_watermark dict records the high watermark for each source
    e.g., high_watermark = {"src1":100,"src2":200}
    src1 or src2 is the canonical_string of partition_locator
    """

    @staticmethod
    def of(
        high_watermark: HighWatermark,
        compacted_delta_locator: DeltaLocator,
        compacted_pyarrow_write_result: PyArrowWriteResult,
        sort_keys_bit_width: int,
        rebase_source_partition_locator: Optional[PartitionLocator] = None,
        manifest_entry_copied_by_reference_ratio: Optional[float] = None,
        compaction_audit_url: Optional[str] = None,
        hash_bucket_count: Optional[int] = None,
        hb_index_to_entry_range: Optional[Dict[int, Tuple[int, int]]] = None,
        compactor_version: Optional[str] = None,
        input_inflation: Optional[float] = None,
        input_average_record_size_bytes: Optional[float] = None,
    ) -> RoundCompletionInfo:

        rci = RoundCompletionInfo()
        rci["highWatermark"] = high_watermark
        rci["compactedDeltaLocator"] = compacted_delta_locator
        rci["compactedPyarrowWriteResult"] = compacted_pyarrow_write_result
        rci["sortKeysBitWidth"] = sort_keys_bit_width
        rci["rebaseSourcePartitionLocator"] = rebase_source_partition_locator
        rci[
            "manifestEntryCopiedByReferenceRatio"
        ] = manifest_entry_copied_by_reference_ratio
        rci["compactionAuditUrl"] = compaction_audit_url
        rci["hashBucketCount"] = hash_bucket_count
        rci["hbIndexToEntryRange"] = hb_index_to_entry_range
        rci["compactorVersion"] = compactor_version
        rci["inputInflation"] = input_inflation
        rci["inputAverageRecordSizeBytes"] = input_average_record_size_bytes
        return rci

    @property
    def high_watermark(self) -> HighWatermark:
        val: Dict[str, Any] = self.get("highWatermark")
        if (
            val is not None
            and isinstance(val, dict)
            and not isinstance(val, HighWatermark)
        ):
            self["highWatermark"] = val = HighWatermark(val)
        return val

    @property
    def compacted_delta_locator(self) -> DeltaLocator:
        val: Dict[str, Any] = self.get("compactedDeltaLocator")
        if val is not None and not isinstance(val, DeltaLocator):
            self["compactedDeltaLocator"] = val = DeltaLocator(val)
        return val

    @property
    def compacted_pyarrow_write_result(self) -> PyArrowWriteResult:
        val: Dict[str, Any] = self.get("compactedPyarrowWriteResult")
        if val is not None and not isinstance(val, PyArrowWriteResult):
            self["compactedPyarrowWriteResult"] = val = PyArrowWriteResult(val)
        return val

    @property
    def sort_keys_bit_width(self) -> int:
        return self["sortKeysBitWidth"]

    @property
    def compaction_audit_url(self) -> Optional[str]:
        return self.get("compactionAuditUrl")

    @property
    def rebase_source_partition_locator(self) -> Optional[PartitionLocator]:
        return self.get("rebaseSourcePartitionLocator")

    @property
    def manifest_entry_copied_by_reference_ratio(self) -> Optional[float]:
        return self["manifestEntryCopiedByReferenceRatio"]

    @property
    def hash_bucket_count(self) -> Optional[int]:
        return self["hashBucketCount"]

    @property
    def hb_index_to_entry_range(self) -> Optional[Dict[int, Tuple[int, int]]]:
        """
        The start index is inclusive and end index is exclusive by default.
        """
        return self["hbIndexToEntryRange"]

    @property
    def compactor_version(self) -> Optional[str]:
        return self.get("compactorVersion")

    @property
    def input_inflation(self) -> Optional[float]:
        return self.get("inputInflation")

    @property
    def input_average_record_size_bytes(self) -> Optional[float]:
        return self.get("inputAverageRecordSizeBytes")



import importlib
from contextlib import nullcontext
import functools
import logging
import ray
import time
import json
from deltacat.aws import s3u as s3_utils
import deltacat
from deltacat import logs
import pyarrow as pa
from deltacat.compute.compactor import (
    PyArrowWriteResult,
    RoundCompletionInfo,
)
from deltacat.storage.model.sort_key import SortKey
from deltacat.compute.compactor.model.dedupe_result import DedupeResult
from deltacat.compute.compactor.model.hash_bucket_result import HashBucketResult
from deltacat.io.object_store import IObjectStore
from deltacat.io.ray_plasma_object_store import RayPlasmaObjectStore
from deltacat.compute.compactor.model.materialize_result import MaterializeResult
from deltacat.compute.stats.models.delta_stats import DeltaStats
from deltacat.storage import (
    Delta,
    DeltaLocator,
    Partition,
    PartitionLocator,
    interface as unimplemented_deltacat_storage,
)
from deltacat.compute.compactor.model.compact_partition_params import (
    CompactPartitionParams,
)
from deltacat.utils.ray_utils.concurrency import (
    invoke_parallel,
    round_robin_options_provider,
)
from deltacat.utils.common import ReadKwargsProvider
from deltacat.utils.ray_utils.runtime import live_node_resource_keys
from deltacat.compute.compactor.steps import dedupe as dd
from deltacat.compute.compactor.steps import hash_bucket as hb
from deltacat.compute.compactor.steps import materialize as mat
from deltacat.compute.compactor.utils import io
from deltacat.compute.compactor.utils import round_completion_file as rcf

from deltacat.types.media import ContentType
from deltacat.utils.placement import PlacementGroupConfig
from typing import List, Set, Optional, Tuple, Dict, Any
from collections import defaultdict
from deltacat.utils.metrics import MetricsConfig
from deltacat.compute.compactor.model.compaction_session_audit_info import (
    CompactionSessionAuditInfo,
)
from deltacat.compute.compactor.model.compactor_version import CompactorVersion
from deltacat.compute.compactor.utils.sort_key import validate_sort_keys
from deltacat.utils.resources import get_current_node_peak_memory_usage_in_bytes


if importlib.util.find_spec("memray"):
    import memray


logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

DEFAULT_DEDUPE_MAX_PARALLELISM_RATIO_ARG: int = 1
DEFAULT_PROPERTIES_ARG: Dict[str, Any] = {}


def check_preconditions(
    source_partition_locator: PartitionLocator,
    destination_partition_locator: PartitionLocator,
    sort_keys: List[SortKey],
    max_records_per_output_file: int,
    new_hash_bucket_count: Optional[int],
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> int:
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}
    assert (
        source_partition_locator.partition_values
        == destination_partition_locator.partition_values
    ), (
        "In-place compaction must use the same partition values for the "
        "source and destination."
    )
    assert (
        max_records_per_output_file >= 1
    ), "Max records per output file must be a positive value"
    if new_hash_bucket_count is not None:
        assert (
            new_hash_bucket_count >= 1
        ), "New hash bucket count must be a positive value"
    return validate_sort_keys(
        source_partition_locator,
        sort_keys,
        deltacat_storage,
        deltacat_storage_kwargs,
        **kwargs,
    )


def compact_partition(
    source_partition_locator: PartitionLocator,
    destination_partition_locator: PartitionLocator,
    primary_keys: Set[str],
    compaction_artifact_s3_bucket: str,
    last_stream_position_to_compact: int,
    *,
    hash_bucket_count: Optional[int] = None,
    sort_keys: List[SortKey] = None,
    records_per_compacted_file: int = 4_000_000,
    input_deltas_stats: Dict[int, DeltaStats] = None,
    min_hash_bucket_chunk_size: int = 0,
    compacted_file_content_type: ContentType = ContentType.PARQUET,
    pg_config: Optional[PlacementGroupConfig] = None,
    schema_on_read: Optional[
        pa.schema
    ] = None,  # TODO (ricmiyam): Remove this and retrieve schema from storage API
    rebase_source_partition_locator: Optional[PartitionLocator] = None,
    rebase_source_partition_high_watermark: Optional[int] = None,
    enable_profiler: Optional[bool] = False,
    metrics_config: Optional[MetricsConfig] = None,
    list_deltas_kwargs: Optional[Dict[str, Any]] = None,
    read_kwargs_provider: Optional[ReadKwargsProvider] = None,
    s3_table_writer_kwargs: Optional[Dict[str, Any]] = None,
    object_store: Optional[IObjectStore] = RayPlasmaObjectStore(),
    s3_client_kwargs: Optional[Dict[str, Any]] = None,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> Optional[str]:
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}
    if not importlib.util.find_spec("memray"):
        logger.info(f"memray profiler not available, disabling all profiling")
        enable_profiler = False

    if s3_client_kwargs is None:
        s3_client_kwargs = {}

    # memray official documentation link:
    # https://bloomberg.github.io/memray/getting_started.html
    with memray.Tracker(
        f"compaction_partition.bin"
    ) if enable_profiler else nullcontext():
        partition = None
        (
            new_partition,
            new_rci,
            new_rcf_partition_locator,
        ) = _execute_compaction_round(
            source_partition_locator,
            destination_partition_locator,
            primary_keys,
            compaction_artifact_s3_bucket,
            last_stream_position_to_compact,
            hash_bucket_count,
            sort_keys,
            records_per_compacted_file,
            input_deltas_stats,
            min_hash_bucket_chunk_size,
            compacted_file_content_type,
            pg_config,
            schema_on_read,
            rebase_source_partition_locator,
            rebase_source_partition_high_watermark,
            enable_profiler,
            metrics_config,
            list_deltas_kwargs,
            read_kwargs_provider,
            s3_table_writer_kwargs,
            object_store,
            s3_client_kwargs,
            deltacat_storage,
            deltacat_storage_kwargs,
            **kwargs,
        )
        if new_partition:
            partition = new_partition

        logger.info(
            f"Partition-{source_partition_locator.partition_values}-> Compaction session data processing completed"
        )
        round_completion_file_s3_url = None
        if partition:
            logger.info(f"Committing compacted partition to: {partition.locator}")
            partition = deltacat_storage.commit_partition(
                partition, **deltacat_storage_kwargs
            )
            logger.info(f"Committed compacted partition: {partition}")

            round_completion_file_s3_url = rcf.write_round_completion_file(
                compaction_artifact_s3_bucket,
                new_rcf_partition_locator,
                new_rci,
                **s3_client_kwargs,
            )
        logger.info(f"Completed compaction session for: {source_partition_locator}")
        return round_completion_file_s3_url


def _execute_compaction_round(
    source_partition_locator: PartitionLocator,
    destination_partition_locator: PartitionLocator,
    primary_keys: Set[str],
    compaction_artifact_s3_bucket: str,
    last_stream_position_to_compact: int,
    hash_bucket_count: Optional[int],
    sort_keys: List[SortKey],
    records_per_compacted_file: int,
    input_deltas_stats: Dict[int, DeltaStats],
    min_hash_bucket_chunk_size: int,
    compacted_file_content_type: ContentType,
    pg_config: Optional[PlacementGroupConfig],
    schema_on_read: Optional[pa.schema],
    rebase_source_partition_locator: Optional[PartitionLocator],
    rebase_source_partition_high_watermark: Optional[int],
    enable_profiler: Optional[bool],
    metrics_config: Optional[MetricsConfig],
    list_deltas_kwargs: Optional[Dict[str, Any]],
    read_kwargs_provider: Optional[ReadKwargsProvider],
    s3_table_writer_kwargs: Optional[Dict[str, Any]],
    object_store: Optional[IObjectStore],
    s3_client_kwargs: Optional[Dict[str, Any]],
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> Tuple[Optional[Partition], Optional[RoundCompletionInfo], Optional[str]]:
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}
    rcf_source_partition_locator = (
        rebase_source_partition_locator
        if rebase_source_partition_locator
        else source_partition_locator
    )
    base_audit_url = rcf_source_partition_locator.path(
        f"s3://{compaction_artifact_s3_bucket}/compaction-audit"
    )
    audit_url = f"{base_audit_url}.json"

    logger.info(f"Compaction audit will be written to {audit_url}")

    compaction_audit = CompactionSessionAuditInfo(
        deltacat.__version__,
        ray.__version__,
        audit_url,
    ).set_compactor_version(CompactorVersion.V1.value)

    compaction_start = time.monotonic()

    if not primary_keys:
        # TODO (pdames): run simple rebatch to reduce all deltas into 1 delta
        #  with normalized manifest entry sizes
        raise NotImplementedError(
            "Compaction only supports tables with 1 or more primary keys"
        )
    if sort_keys is None:
        sort_keys = []
    # TODO (pdames): detect and handle schema evolution (at least ensure that
    #  we don't recompact simple backwards-compatible changes like constraint
    #  widening and null column additions).
    # TODO (pdames): detect and optimize in-place compaction

    # check preconditions before doing any computationally expensive work
    bit_width_of_sort_keys = check_preconditions(
        source_partition_locator,
        destination_partition_locator,
        sort_keys,
        records_per_compacted_file,
        hash_bucket_count,
        deltacat_storage,
        deltacat_storage_kwargs,
        **kwargs,
    )

    # sort primary keys to produce the same pk digest regardless of input order
    primary_keys = sorted(primary_keys)

    node_resource_keys = None
    if pg_config:  # use resource in each placement group
        cluster_resources = pg_config.resource
        cluster_cpus = cluster_resources["CPU"]
    else:  # use all cluster resource
        cluster_resources = ray.cluster_resources()
        logger.info(f"Total cluster resources: {cluster_resources}")
        logger.info(f"Available cluster resources: {ray.available_resources()}")
        cluster_cpus = int(cluster_resources["CPU"])
        logger.info(f"Total cluster CPUs: {cluster_cpus}")
        node_resource_keys = live_node_resource_keys()
        logger.info(
            f"Found {len(node_resource_keys)} live cluster nodes: "
            f"{node_resource_keys}"
        )

    compaction_audit.set_cluster_cpu_max(cluster_cpus)
    # create a remote options provider to round-robin tasks across all nodes or allocated bundles
    logger.info(f"Setting round robin scheduling with node id:{node_resource_keys}")
    round_robin_opt_provider = functools.partial(
        round_robin_options_provider,
        resource_keys=node_resource_keys,
        pg_config=pg_config.opts if pg_config else None,
    )

    # set max task parallelism equal to total cluster CPUs...
    # we assume here that we're running on a fixed-size cluster - this
    # assumption could be removed but we'd still need to know the maximum
    # "safe" number of parallel tasks that our autoscaling cluster could handle
    max_parallelism = int(cluster_cpus)
    logger.info(f"Max parallelism: {max_parallelism}")

    # read the results from any previously completed compaction round
    round_completion_info = None
    if not rebase_source_partition_locator:
        round_completion_info = rcf.read_round_completion_file(
            compaction_artifact_s3_bucket, source_partition_locator, **s3_client_kwargs
        )
        if not round_completion_info:
            logger.info(
                f"Both rebase partition and round completion file not found. Performing an entire backfill on source."
            )
        logger.info(f"Round completion file: {round_completion_info}")

    enable_manifest_entry_copy_by_reference = (
        False if rebase_source_partition_locator else True
    )
    logger.info(
        f"Enable manifest entry copy by reference is set to: {enable_manifest_entry_copy_by_reference}"
    )

    # discover input delta files
    # For rebase:
    # Copy the old compacted table to a new destination, plus any new deltas from rebased source

    # For incremental compaction:
    # Input deltas from two sources: One: new delta; Two: compacted table

    high_watermark = (
        round_completion_info.high_watermark if round_completion_info else None
    )

    delta_discovery_start = time.monotonic()
    (
        input_deltas,
        previous_last_stream_position_compacted_on_destination_table,
    ) = io.discover_deltas(
        source_partition_locator,
        high_watermark,
        last_stream_position_to_compact,
        destination_partition_locator,
        rebase_source_partition_locator,
        rebase_source_partition_high_watermark,
        deltacat_storage,
        deltacat_storage_kwargs,
        list_deltas_kwargs,
    )

    delta_discovery_end = time.monotonic()
    compaction_audit.set_delta_discovery_time_in_seconds(
        delta_discovery_end - delta_discovery_start
    )

    s3_utils.upload(
        compaction_audit.audit_url,
        str(json.dumps(compaction_audit)),
        **s3_client_kwargs,
    )

    if not input_deltas:
        logger.info("No input deltas found to compact.")
        return None, None, None

    # limit the input deltas to fit on this cluster and convert them to
    # annotated deltas of equivalent size for easy parallel distribution

    (
        uniform_deltas,
        hash_bucket_count,
        last_stream_position_compacted,
        require_multiple_rounds,
    ) = (
        io.fit_input_deltas(
            input_deltas,
            cluster_resources,
            compaction_audit,
            hash_bucket_count,
            deltacat_storage=deltacat_storage,
            deltacat_storage_kwargs=deltacat_storage_kwargs,
            **kwargs,
        )
        if input_deltas_stats is None
        else io.limit_input_deltas(
            input_deltas,
            cluster_resources,
            hash_bucket_count,
            min_hash_bucket_chunk_size,
            compaction_audit=compaction_audit,
            input_deltas_stats=input_deltas_stats,
            deltacat_storage=deltacat_storage,
            deltacat_storage_kwargs=deltacat_storage_kwargs,
            **kwargs,
        )
    )

    compaction_audit.set_uniform_deltas_created(len(uniform_deltas))

    assert hash_bucket_count is not None and hash_bucket_count > 0, (
        f"Expected hash bucket count to be a positive integer, but found "
        f"`{hash_bucket_count}`"
    )
    # parallel step 1:
    # group like primary keys together by hashing them into buckets
    if not round_completion_info and rebase_source_partition_locator:
        # generate a rc_info for hb and materialize to tell whether a delta is src or destination in case of rebase
        dest_delta_locator = DeltaLocator.of(
            partition_locator=rebase_source_partition_locator, stream_position=None
        )
        round_completion_info = RoundCompletionInfo.of(
            None, dest_delta_locator, None, 0, None
        )

    if require_multiple_rounds:
        logger.info(
            f"Compaction can not be completed in one round. Either increase cluster size or decrease input"
        )
        raise AssertionError(
            "Multiple rounds are not supported. Please increase the cluster size and run again."
        )
    hb_start = time.monotonic()
    hb_tasks_pending = invoke_parallel(
        items=uniform_deltas,
        ray_task=hb.hash_bucket,
        max_parallelism=max_parallelism,
        options_provider=round_robin_opt_provider,
        round_completion_info=round_completion_info,
        primary_keys=primary_keys,
        sort_keys=sort_keys,
        num_buckets=hash_bucket_count,
        num_groups=max_parallelism,
        enable_profiler=enable_profiler,
        metrics_config=metrics_config,
        read_kwargs_provider=read_kwargs_provider,
        object_store=object_store,
        deltacat_storage=deltacat_storage,
        deltacat_storage_kwargs=deltacat_storage_kwargs,
        **kwargs,
    )
    hb_invoke_end = time.monotonic()

    logger.info(f"Getting {len(hb_tasks_pending)} hash bucket results...")
    hb_results: List[HashBucketResult] = ray.get(hb_tasks_pending)
    logger.info(f"Got {len(hb_results)} hash bucket results.")
    hb_end = time.monotonic()
    hb_results_retrieved_at = time.time()

    telemetry_time_hb = compaction_audit.save_step_stats(
        CompactionSessionAuditInfo.HASH_BUCKET_STEP_NAME,
        hb_results,
        hb_results_retrieved_at,
        hb_invoke_end - hb_start,
        hb_end - hb_start,
    )

    s3_utils.upload(
        compaction_audit.audit_url,
        str(json.dumps(compaction_audit)),
        **s3_client_kwargs,
    )

    all_hash_group_idx_to_obj_id = defaultdict(list)
    for hb_result in hb_results:
        for hash_group_index, object_id in enumerate(
            hb_result.hash_bucket_group_to_obj_id
        ):
            if object_id:
                all_hash_group_idx_to_obj_id[hash_group_index].append(object_id)
    hash_group_count = len(all_hash_group_idx_to_obj_id)
    logger.info(f"Hash bucket groups created: {hash_group_count}")
    total_hb_record_count = sum([hb_result.hb_record_count for hb_result in hb_results])
    logger.info(
        f"Got {total_hb_record_count} hash bucket records from hash bucketing step..."
    )

    compaction_audit.set_input_records(total_hb_record_count.item())
    # TODO (pdames): when resources are freed during the last round of hash
    #  bucketing, start running dedupe tasks that read existing dedupe
    #  output from S3 then wait for hash bucketing to finish before continuing

    # create a new stream for this round
    compacted_stream_locator = destination_partition_locator.stream_locator
    stream = deltacat_storage.get_stream(
        compacted_stream_locator.namespace,
        compacted_stream_locator.table_name,
        compacted_stream_locator.table_version,
        **deltacat_storage_kwargs,
    )
    partition = deltacat_storage.stage_partition(
        stream,
        destination_partition_locator.partition_values,
        **deltacat_storage_kwargs,
    )
    new_compacted_partition_locator = partition.locator
    # parallel step 2:
    # discover records with duplicate primary keys in each hash bucket, and
    # identify the index of records to keep or drop based on sort keys
    num_materialize_buckets = max_parallelism
    logger.info(f"Materialize Bucket Count: {num_materialize_buckets}")

    dedupe_start = time.monotonic()
    dd_max_parallelism = int(
        max_parallelism
        * kwargs.get(
            "dd_max_parallelism_ratio", DEFAULT_DEDUPE_MAX_PARALLELISM_RATIO_ARG
        )
    )
    logger.info(
        f"dd max_parallelism is set to {dd_max_parallelism}, max_parallelism is {max_parallelism}"
    )
    dd_tasks_pending = invoke_parallel(
        items=all_hash_group_idx_to_obj_id.values(),
        ray_task=dd.dedupe,
        max_parallelism=dd_max_parallelism,
        options_provider=round_robin_opt_provider,
        kwargs_provider=lambda index, item: {
            "dedupe_task_index": index,
            "object_ids": item,
        },
        sort_keys=sort_keys,
        num_materialize_buckets=num_materialize_buckets,
        enable_profiler=enable_profiler,
        metrics_config=metrics_config,
        object_store=object_store,
    )

    dedupe_invoke_end = time.monotonic()
    logger.info(f"Getting {len(dd_tasks_pending)} dedupe results...")
    dd_results: List[DedupeResult] = ray.get(dd_tasks_pending)
    logger.info(f"Got {len(dd_results)} dedupe results.")

    # we use time.time() here because time.monotonic() has no reference point
    # whereas time.time() measures epoch seconds. Hence, it will be reasonable
    # to compare time.time()s captured in different nodes.
    dedupe_results_retrieved_at = time.time()
    dedupe_end = time.monotonic()

    total_dd_record_count = sum([ddr.deduped_record_count for ddr in dd_results])
    logger.info(f"Deduped {total_dd_record_count} records...")

    telemetry_time_dd = compaction_audit.save_step_stats(
        CompactionSessionAuditInfo.DEDUPE_STEP_NAME,
        dd_results,
        dedupe_results_retrieved_at,
        dedupe_invoke_end - dedupe_start,
        dedupe_end - dedupe_start,
    )

    compaction_audit.set_records_deduped(total_dd_record_count.item())
    all_mat_buckets_to_obj_id = defaultdict(list)
    for dd_result in dd_results:
        for (
            bucket_idx,
            dd_task_index_and_object_id_tuple,
        ) in dd_result.mat_bucket_idx_to_obj_id.items():
            all_mat_buckets_to_obj_id[bucket_idx].append(
                dd_task_index_and_object_id_tuple
            )
    logger.info(f"Getting {len(dd_tasks_pending)} dedupe result stat(s)...")
    logger.info(f"Materialize buckets created: " f"{len(all_mat_buckets_to_obj_id)}")

    compaction_audit.set_materialize_buckets(len(all_mat_buckets_to_obj_id))
    # TODO(pdames): when resources are freed during the last round of deduping
    #  start running materialize tasks that read materialization source file
    #  tables from S3 then wait for deduping to finish before continuing

    # TODO(pdames): balance inputs to materialization tasks to ensure that each
    #  task has an approximately equal amount of input to materialize

    # TODO(pdames): garbage collect hash bucket output since it's no longer
    #  needed

    # parallel step 3:
    # materialize records to keep by index

    s3_utils.upload(
        compaction_audit.audit_url,
        str(json.dumps(compaction_audit)),
        **s3_client_kwargs,
    )

    materialize_start = time.monotonic()
    mat_tasks_pending = invoke_parallel(
        items=all_mat_buckets_to_obj_id.items(),
        ray_task=mat.materialize,
        max_parallelism=max_parallelism,
        options_provider=round_robin_opt_provider,
        kwargs_provider=lambda index, mat_bucket_index_to_obj_id: {
            "mat_bucket_index": mat_bucket_index_to_obj_id[0],
            "dedupe_task_idx_and_obj_id_tuples": mat_bucket_index_to_obj_id[1],
        },
        schema=schema_on_read,
        round_completion_info=round_completion_info,
        source_partition_locator=source_partition_locator,
        partition=partition,
        enable_manifest_entry_copy_by_reference=enable_manifest_entry_copy_by_reference,
        max_records_per_output_file=records_per_compacted_file,
        compacted_file_content_type=compacted_file_content_type,
        enable_profiler=enable_profiler,
        metrics_config=metrics_config,
        read_kwargs_provider=read_kwargs_provider,
        s3_table_writer_kwargs=s3_table_writer_kwargs,
        object_store=object_store,
        deltacat_storage=deltacat_storage,
        deltacat_storage_kwargs=deltacat_storage_kwargs,
    )

    materialize_invoke_end = time.monotonic()

    logger.info(f"Getting {len(mat_tasks_pending)} materialize result(s)...")
    mat_results: List[MaterializeResult] = ray.get(mat_tasks_pending)

    logger.info(f"Got {len(mat_results)} materialize result(s).")

    materialize_end = time.monotonic()
    materialize_results_retrieved_at = time.time()

    telemetry_time_materialize = compaction_audit.save_step_stats(
        CompactionSessionAuditInfo.MATERIALIZE_STEP_NAME,
        mat_results,
        materialize_results_retrieved_at,
        materialize_invoke_end - materialize_start,
        materialize_end - materialize_start,
    )

    mat_results = sorted(mat_results, key=lambda m: m.task_index)
    deltas = [m.delta for m in mat_results]

    # Note: An appropriate last stream position must be set
    # to avoid correctness issue.
    merged_delta = Delta.merge_deltas(
        deltas,
        stream_position=last_stream_position_to_compact,
    )

    record_info_msg = (
        f"Hash bucket records: {total_hb_record_count},"
        f" Deduped records: {total_dd_record_count}, "
        f" Materialized records: {merged_delta.meta.record_count}"
    )
    logger.info(record_info_msg)

    assert (
        total_hb_record_count - total_dd_record_count == merged_delta.meta.record_count
    ), (
        f"Number of hash bucket records minus the number of deduped records"
        f" does not match number of materialized records.\n"
        f" {record_info_msg}"
    )
    compacted_delta = deltacat_storage.commit_delta(
        merged_delta,
        properties=kwargs.get("properties", DEFAULT_PROPERTIES_ARG),
        **deltacat_storage_kwargs,
    )
    logger.info(f"Committed compacted delta: {compacted_delta}")

    compaction_end = time.monotonic()
    compaction_audit.set_compaction_time_in_seconds(compaction_end - compaction_start)

    new_compacted_delta_locator = DeltaLocator.of(
        new_compacted_partition_locator,
        compacted_delta.stream_position,
    )

    last_rebase_source_partition_locator = rebase_source_partition_locator or (
        round_completion_info.rebase_source_partition_locator
        if round_completion_info
        else None
    )

    pyarrow_write_result = PyArrowWriteResult.union(
        [m.pyarrow_write_result for m in mat_results]
    )

    session_peak_memory = get_current_node_peak_memory_usage_in_bytes()
    compaction_audit.set_peak_memory_used_bytes_by_compaction_session_process(
        session_peak_memory
    )

    compaction_audit.save_round_completion_stats(
        mat_results, telemetry_time_hb + telemetry_time_dd + telemetry_time_materialize
    )

    s3_utils.upload(
        compaction_audit.audit_url,
        str(json.dumps(compaction_audit)),
        **s3_client_kwargs,
    )

    new_round_completion_info = RoundCompletionInfo.of(
        last_stream_position_compacted,
        new_compacted_delta_locator,
        pyarrow_write_result,
        bit_width_of_sort_keys,
        last_rebase_source_partition_locator,
        compaction_audit.untouched_file_ratio,
        audit_url,
        hash_bucket_count,
        None,
        CompactorVersion.V1.value,
    )

    logger.info(
        f"partition-{source_partition_locator.partition_values},"
        f"compacted at: {last_stream_position_compacted},"
        f"last position: {last_stream_position_to_compact}"
    )

    return (
        partition,
        new_round_completion_info,
        rcf_source_partition_locator,
    )


def compact_partition_from_request(
    compact_partition_params: CompactPartitionParams,
    *compact_partition_pos_args,
) -> Optional[str]:
    """
    Wrapper for compact_partition that allows for the compact_partition parameters to be
    passed in as a custom dictionary-like CompactPartitionParams object along with any compact_partition positional arguments.
    :param compact_partition_params:
    """
    return compact_partition(*compact_partition_pos_args, **compact_partition_params)



import unittest
from unittest.mock import MagicMock
import pyarrow as pa
from deltacat.compute.compactor.steps.repartition import repartition_range
from deltacat.types.media import ContentType
from deltacat.compute.compactor.model.repartition_result import RepartitionResult
from deltacat.storage import (
    PartitionLocator,
)

"""
Summary of Test Cases:
0. Given empty ranges, error should be raised
1. Given one value in ranges, e.g., [1678665487112747], Two range deltas should be produced
2. Given two values in ranges, e.g., [1678665487112747, 1678665487112999], three range deltas should be produced
3. No enough records exist for all ranges, i.e., some range will have empty file, such that number of returned range deltas should be less
4. column doens't exist in any table, error should be raised
5. column exists in some table, but not all, error should be raised
6. Given ranges is unsorted , e.g., [1678665487112747, 1678665487112745, 1678665487112748]
7. Given ranges may have same values, e.g., [1678665487112747, 1678665487112747]
8. Ranges with pre-dfined inf, e.g., [1678665487112747, inf]
"""


class TestRepartitionRange(unittest.TestCase):
    def setUp(self):
        self.tables = [
            pa.table(
                {
                    "last_updated": [
                        1678665487112745,
                        1678665487112746,
                        1678665487112747,
                        1678665487112748,
                    ]
                }
            ),
            pa.table(
                {
                    "last_updated": [
                        1678665487112748,
                        1678665487112749,
                        1678665487112750,
                        1678665487112751,
                    ]
                }
            ),
        ]
        self.destination_partition: PartitionLocator = MagicMock()
        self.repartition_args = {"column": "last_updated", "ranges": [1678665487112747]}
        self.max_records_per_output_file = 2
        self.s3_table_writer_kwargs = {}
        self.repartitioned_file_content_type = ContentType.PARQUET
        self.deltacat_storage = MagicMock()
        self.deltacat_storage_kwargs = MagicMock()

    def test_repartition_range(self):
        result = repartition_range(
            self.tables,
            self.destination_partition,
            self.repartition_args,
            self.max_records_per_output_file,
            self.s3_table_writer_kwargs,
            self.repartitioned_file_content_type,
            self.deltacat_storage,
            self.deltacat_storage_kwargs,
        )
        # Assert that a RepartitionResult object is returned
        self.assertIsInstance(result, RepartitionResult)

        # Assert that the correct number of range_deltas was produced
        self.assertEqual(
            len(result.range_deltas), len(self.repartition_args["ranges"]) + 1
        )

        # Assert that the function called the deltacat_storage.stage_delta method the correct number of times
        self.assertEqual(
            self.deltacat_storage.stage_delta.call_count,
            len(self.repartition_args["ranges"]) + 1,
        )

    def test_repartition_range_nonexistent_column(self):
        self.repartition_args["column"] = "nonexistent_column"
        with self.assertRaises(ValueError):
            repartition_range(
                self.tables,
                self.destination_partition,
                self.repartition_args,
                self.max_records_per_output_file,
                self.s3_table_writer_kwargs,
                self.repartitioned_file_content_type,
                self.deltacat_storage,
                self.deltacat_storage_kwargs,
            )

    def test_empty_ranges(self):
        self.repartition_args["ranges"] = []
        with self.assertRaises(ValueError):
            repartition_range(
                self.tables,
                self.destination_partition,
                self.repartition_args,
                self.max_records_per_output_file,
                self.s3_table_writer_kwargs,
                self.repartitioned_file_content_type,
                self.deltacat_storage,
                self.deltacat_storage_kwargs,
            )

    def test_one_value_in_ranges(self):
        self.repartition_args["ranges"] = [1678665487112747]
        result = repartition_range(
            self.tables,
            self.destination_partition,
            self.repartition_args,
            self.max_records_per_output_file,
            self.s3_table_writer_kwargs,
            self.repartitioned_file_content_type,
            self.deltacat_storage,
            self.deltacat_storage_kwargs,
        )
        self.assertEqual(len(result.range_deltas), 2)

    def test_two_values_in_ranges(self):
        self.repartition_args["ranges"] = [1678665487112747, 1678665487112749]
        result = repartition_range(
            self.tables,
            self.destination_partition,
            self.repartition_args,
            self.max_records_per_output_file,
            self.s3_table_writer_kwargs,
            self.repartitioned_file_content_type,
            self.deltacat_storage,
            self.deltacat_storage_kwargs,
        )
        self.assertEqual(len(result.range_deltas), 3)

    def test_not_enough_records_for_all_ranges(self):
        reduced_tables = [self.tables[0]]  # use only the first table
        self.repartition_args["ranges"] = [1678665487112749, 1678665487112999]
        result = repartition_range(
            reduced_tables,
            self.destination_partition,
            self.repartition_args,
            self.max_records_per_output_file,
            self.s3_table_writer_kwargs,
            self.repartitioned_file_content_type,
            self.deltacat_storage,
            self.deltacat_storage_kwargs,
        )
        self.assertLess(len(result.range_deltas), 2)

    def test_column_does_not_exist_in_all_tables(self):
        self.tables.append(pa.table({"other_column": [1, 2, 3]}))
        with self.assertRaises(ValueError):
            repartition_range(
                self.tables,
                self.destination_partition,
                self.repartition_args,
                self.max_records_per_output_file,
                self.s3_table_writer_kwargs,
                self.repartitioned_file_content_type,
                self.deltacat_storage,
                self.deltacat_storage_kwargs,
            )

    def test_unsorted_ranges(self):
        self.repartition_args["ranges"] = [
            1678665487112747,
            1678665487112745,
            1678665487112748,
        ]
        result = repartition_range(
            self.tables,
            self.destination_partition,
            self.repartition_args,
            self.max_records_per_output_file,
            self.s3_table_writer_kwargs,
            self.repartitioned_file_content_type,
            self.deltacat_storage,
            self.deltacat_storage_kwargs,
        )
        self.assertEqual(len(result.range_deltas), 4)

    def test_same_values_in_ranges(self):
        self.repartition_args["ranges"] = [1678665487112747, 1678665487112747]
        result = repartition_range(
            self.tables,
            self.destination_partition,
            self.repartition_args,
            self.max_records_per_output_file,
            self.s3_table_writer_kwargs,
            self.repartitioned_file_content_type,
            self.deltacat_storage,
            self.deltacat_storage_kwargs,
        )
        self.assertEqual(len(result.range_deltas), 2)

    def test_ranges_with_inf(self):
        self.repartition_args["ranges"] = [1678665487112747, float("inf")]

        self.assertRaises(
            pa.lib.ArrowInvalid,
            lambda: repartition_range(
                self.tables,
                self.destination_partition,
                self.repartition_args,
                self.max_records_per_output_file,
                self.s3_table_writer_kwargs,
                self.repartitioned_file_content_type,
                self.deltacat_storage,
            ),
        )

    def test_null_rows_are_not_dropped(self):
        # Add null value to the first table
        tables_with_null = [
            pa.table(
                {
                    "last_updated": [
                        None,
                        1678665487112746,
                        1678665487112747,
                        1678665487112748,
                    ]
                }
            ),
            self.tables[1],
        ]

        result = repartition_range(
            tables_with_null,
            self.destination_partition,
            self.repartition_args,
            self.max_records_per_output_file,
            self.s3_table_writer_kwargs,
            self.repartitioned_file_content_type,
            self.deltacat_storage,
            self.deltacat_storage_kwargs,
        )

        # Assuming range_deltas is a list of DataFrames,
        # check that the first DataFrame has the null value in the 'last_updated' column
        # This may need to be adjusted depending on the actual structure of range_deltas
        self.assertEqual(len(result.range_deltas), 2)


if __name__ == "__main__":
    unittest.main()



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations
from enum import Enum
from typing import Any, Dict, List, Optional, Set
import datetime as dt
from boto3.resources.base import ServiceResource
from datetime import timezone

from deltacat.tests.compute.test_util_constant import (
    TEST_S3_RCF_BUCKET_NAME,
)
from deltacat.tests.compute.test_util_constant import (
    BASE_TEST_SOURCE_NAMESPACE,
    BASE_TEST_SOURCE_TABLE_NAME,
    BASE_TEST_SOURCE_TABLE_VERSION,
    BASE_TEST_DESTINATION_NAMESPACE,
    BASE_TEST_DESTINATION_TABLE_NAME,
    BASE_TEST_DESTINATION_TABLE_VERSION,
    REBASING_NAMESPACE,
    REBASING_TABLE_NAME,
    REBASING_TABLE_VERSION,
)


class PartitionKeyType(str, Enum):
    INT = "int"
    STRING = "string"
    TIMESTAMP = "timestamp"


class PartitionKey(dict):
    @staticmethod
    def of(key_name: str, key_type: PartitionKeyType) -> PartitionKey:
        return PartitionKey({"keyName": key_name, "keyType": key_type.value})

    @property
    def key_name(self) -> str:
        return self["keyName"]

    @property
    def key_type(self) -> PartitionKeyType:
        key_type = self["keyType"]
        return None if key_type is None else PartitionKeyType(key_type)


"""
UTILS
"""


def _create_table(
    namespace: str,
    table_name: str,
    table_version: str,
    primary_keys: Set[str],
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    ds_mock_kwargs: Optional[Dict[str, Any]],
):
    import deltacat.tests.local_deltacat_storage as ds
    from deltacat.types.media import ContentType

    ds.create_namespace(namespace, {}, **ds_mock_kwargs)
    ds.create_table_version(
        namespace,
        table_name,
        table_version,
        primary_key_column_names=list(primary_keys),
        sort_keys=sort_keys,
        partition_keys=partition_keys,
        supported_content_types=[ContentType.PARQUET],
        **ds_mock_kwargs,
    )
    return namespace, table_name, table_version


def create_src_table(
    primary_keys: Set[str],
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    ds_mock_kwargs: Optional[Dict[str, Any]],
):
    source_namespace: str = BASE_TEST_SOURCE_NAMESPACE
    source_table_name: str = BASE_TEST_SOURCE_TABLE_NAME
    source_table_version: str = BASE_TEST_SOURCE_TABLE_VERSION
    return _create_table(
        source_namespace,
        source_table_name,
        source_table_version,
        primary_keys,
        sort_keys,
        partition_keys,
        ds_mock_kwargs,
    )


def create_destination_table(
    primary_keys: Set[str],
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    ds_mock_kwargs: Optional[Dict[str, Any]],
):
    destination_namespace: str = BASE_TEST_DESTINATION_NAMESPACE
    destination_table_name: str = BASE_TEST_DESTINATION_TABLE_NAME
    destination_table_version: str = BASE_TEST_DESTINATION_TABLE_VERSION
    return _create_table(
        destination_namespace,
        destination_table_name,
        destination_table_version,
        primary_keys,
        sort_keys,
        partition_keys,
        ds_mock_kwargs,
    )


def create_rebase_table(
    primary_keys: Set[str],
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    ds_mock_kwargs: Optional[Dict[str, Any]],
):
    rebasing_namespace = REBASING_NAMESPACE
    rebasing_table_name = REBASING_TABLE_NAME
    rebasing_table_version = REBASING_TABLE_VERSION
    return _create_table(
        rebasing_namespace,
        rebasing_table_name,
        rebasing_table_version,
        primary_keys,
        sort_keys,
        partition_keys,
        ds_mock_kwargs,
    )


def get_rcf(s3_resource, rcf_file_s3_uri: str):
    from deltacat.tests.test_utils.utils import read_s3_contents
    from deltacat.compute.compactor import (
        RoundCompletionInfo,
    )

    _, rcf_object_key = rcf_file_s3_uri.rsplit("/", 1)
    rcf_file_output: Dict[str, Any] = read_s3_contents(
        s3_resource, TEST_S3_RCF_BUCKET_NAME, rcf_object_key
    )
    return RoundCompletionInfo(**rcf_file_output)


def get_compacted_delta_locator_from_rcf(
    s3_resource: ServiceResource, rcf_file_s3_uri: str
):
    from deltacat.storage import DeltaLocator
    from deltacat.compute.compactor import (
        RoundCompletionInfo,
    )

    round_completion_info: RoundCompletionInfo = get_rcf(s3_resource, rcf_file_s3_uri)

    compacted_delta_locator: DeltaLocator = (
        round_completion_info.compacted_delta_locator
    )
    return compacted_delta_locator


def offer_iso8601_timestamp_list(
    periods: int,
    unit_of_time: str,
    end_time=dt.datetime(2023, 5, 3, 10, 0, 0, 0, tzinfo=timezone.utc),
) -> List[str]:
    """
    Returns a list of ISO 8601 timestamps, each periods units of time before the start time.

    Args:
    periods: The number of timestamps to return.
    unit_of_time: The unit of time to use for the timestamps. Must be one of "seconds", "minutes", "hours", "days", or "weeks".
    end_time: The end time for the timestamps. Defaults to 2023-05-03T10:00:00Z.

    Returns:
    A list of ISO 8601 timestamps, each periods units of time before the start time.

    Raises:
    ValueError: If the unit_of_time argument is not one of "seconds", "minutes", "hours", "days", or "weeks".
    """
    import datetime as dt

    UTC_ISO_8601_FORMAT_WITHOUT_MILLIS = "%Y-%m-%dT%H:%M:%SZ"  # '2018-09-05T14:09:03Z'

    acceptable_units_of_time = ["seconds", "minutes", "hours", "days", "weeks"]
    if unit_of_time not in acceptable_units_of_time:
        raise ValueError(
            f"unit_of_time {unit_of_time} is not supported. Please use one of these time units: {acceptable_units_of_time}"
        )
    res = []
    for i in range(periods):
        kwarg = {unit_of_time: i}
        res.append(
            (end_time - dt.timedelta(**kwarg)).strftime(
                UTC_ISO_8601_FORMAT_WITHOUT_MILLIS
            )
        )
    return res



import ray
from ray import cloudpickle
from deltacat.io.object_store import IObjectStore
from typing import Any, List


class RayPlasmaObjectStore(IObjectStore):
    """
    An implementation of object store that uses Ray plasma object store.
    """

    def put_many(self, objects: List[object], *args, **kwargs) -> List[Any]:
        result = []
        for obj in objects:
            object_ref = ray.put(obj)
            pickled = cloudpickle.dumps(object_ref)
            result.append(pickled)

        return result

    def get_many(self, refs: List[Any], *args, **kwargs) -> List[object]:
        loaded_refs = [cloudpickle.loads(obj_id) for obj_id in refs]
        return ray.get(loaded_refs)



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Any, Dict, Optional
import numpy as np

from deltacat.compute.compactor.model.pyarrow_write_result import PyArrowWriteResult
from deltacat.storage import Delta


class MaterializeResult(dict):
    @staticmethod
    def of(
        delta: Delta,
        task_index: int,
        pyarrow_write_result: PyArrowWriteResult,
        referenced_pyarrow_write_result: Optional[PyArrowWriteResult] = None,
        peak_memory_usage_bytes: Optional[np.double] = None,
        telemetry_time_in_seconds: Optional[np.double] = None,
        task_completed_at: Optional[np.double] = None,
    ) -> MaterializeResult:
        materialize_result = MaterializeResult()
        materialize_result["delta"] = delta
        materialize_result["taskIndex"] = task_index
        materialize_result["paWriteResult"] = pyarrow_write_result
        materialize_result["referencedPaWriteResult"] = referenced_pyarrow_write_result
        materialize_result["peakMemoryUsageBytes"] = peak_memory_usage_bytes
        materialize_result["telemetryTimeInSeconds"] = telemetry_time_in_seconds
        materialize_result["taskCompletedAt"] = task_completed_at
        return materialize_result

    @property
    def delta(self) -> Delta:
        val: Dict[str, Any] = self.get("delta")
        if val is not None and not isinstance(val, Delta):
            self["delta"] = val = Delta(val)
        return val

    @property
    def task_index(self) -> int:
        return self["taskIndex"]

    @property
    def peak_memory_usage_bytes(self) -> Optional[np.double]:
        return self["peakMemoryUsageBytes"]

    @property
    def telemetry_time_in_seconds(self) -> Optional[np.double]:
        return self["telemetryTimeInSeconds"]

    @property
    def pyarrow_write_result(self) -> PyArrowWriteResult:
        val: Dict[str, Any] = self.get("paWriteResult")
        if val is not None and not isinstance(val, PyArrowWriteResult):
            self["paWriteResult"] = val = PyArrowWriteResult(val)
        return val

    @property
    def count_of_src_dfl_not_touched(self) -> int:
        return self["countOfSrcFileNotTouched"]

    @property
    def referenced_pyarrow_write_result(self) -> PyArrowWriteResult:
        val: Dict[str, Any] = self.get("referencedPaWriteResult")
        if val is not None and not isinstance(val, PyArrowWriteResult):
            self["referencedPaWriteResult"] = val = PyArrowWriteResult(val)

        return val

    @property
    def task_completed_at(self) -> Optional[np.double]:
        return self["taskCompletedAt"]



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

TEST_S3_RCF_BUCKET_NAME = "test-compaction-artifacts-bucket"
# REBASE  src = spark compacted table to create an initial version of ray compacted table
BASE_TEST_SOURCE_NAMESPACE = "source_test_namespace"
BASE_TEST_SOURCE_TABLE_NAME = "source_test_table"
BASE_TEST_SOURCE_TABLE_VERSION = "1"

BASE_TEST_DESTINATION_NAMESPACE = "destination_test_namespace"
BASE_TEST_DESTINATION_TABLE_NAME = "destination_test_table"
BASE_TEST_DESTINATION_TABLE_VERSION = "1"

REBASING_NAMESPACE = "compacted"
REBASING_TABLE_NAME = "rebase_test_table"
REBASING_TABLE_VERSION = "1"
REBASING_NAME_SUFFIX = "_compacted"

RAY_COMPACTED_NAMESPACE = "compacted_ray"
RAY_COMPACTED_NAME_SUFFIX = "_compacted_ray"

DEFAULT_HASH_BUCKET_COUNT: int = 3

DEFAULT_MAX_RECORDS_PER_FILE: int = 4_000_000

DEFAULT_NUM_WORKERS = 1
DEFAULT_WORKER_INSTANCE_CPUS = 1



import logging
from deltacat import logs
from deltacat.storage import (
    Delta,
    interface as unimplemented_deltacat_storage,
)
from typing import Dict, Optional, Any
from deltacat.types.media import TableType, StorageType
from deltacat.types.media import ContentType
from deltacat.types.partial_download import PartialParquetParameters

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def append_content_type_params(
    delta: Delta,
    entry_index: Optional[int] = None,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = {},
) -> None:

    if delta.meta.content_type != ContentType.PARQUET.value:
        logger.info(
            f"Delta with locator {delta.locator} is not a parquet delta, "
            "skipping appending content type parameters."
        )
        return

    manifest_entries = delta.manifest.entries
    ordered_pq_meta = []

    if entry_index is not None:
        manifest_entries = [delta.manifest.entries[entry_index]]

        pq_file = deltacat_storage.download_delta_manifest_entry(
            delta,
            entry_index=entry_index,
            table_type=TableType.PYARROW_PARQUET,
            **deltacat_storage_kwargs,
        )

        partial_file_meta = PartialParquetParameters.of(pq_metadata=pq_file.metadata)
        ordered_pq_meta.append(partial_file_meta)

    else:
        pq_files = deltacat_storage.download_delta(
            delta,
            table_type=TableType.PYARROW_PARQUET,
            storage_type=StorageType.LOCAL,
            **deltacat_storage_kwargs,
        )

        assert len(pq_files) == len(
            manifest_entries
        ), f"Expected {len(manifest_entries)} pq files, got {len(pq_files)}"

        ordered_pq_meta = [
            PartialParquetParameters.of(pq_metadata=pq_file.metadata)
            for pq_file in pq_files
        ]

    for entry_index, entry in enumerate(manifest_entries):
        if not entry.meta.content_type_parameters:
            entry.meta.content_type_parameters = []

        entry.meta.content_type_parameters.append(ordered_pq_meta[entry_index])



import logging

import deltacat.logs  # noqa: F401
from deltacat.catalog.delegate import (
    alter_namespace,
    alter_table,
    create_namespace,
    create_table,
    default_namespace,
    drop_namespace,
    drop_table,
    get_namespace,
    get_table,
    list_namespaces,
    list_tables,
    namespace_exists,
    read_table,
    refresh_table,
    rename_table,
    table_exists,
    truncate_table,
    write_to_table,
)
from deltacat.catalog.model.catalog import (  # noqa: F401
    Catalog,
    Catalogs,
    all_catalogs,
    init,
)
from deltacat.catalog.model.table_definition import TableDefinition
from deltacat.storage import (
    DistributedDataset,
    LifecycleState,
    ListResult,
    LocalDataset,
    LocalTable,
    Namespace,
    SchemaConsistencyType,
    SortKey,
    SortOrder,
)
from deltacat.types.media import ContentEncoding, ContentType, TableType
from deltacat.types.tables import TableWriteMode

deltacat.logs.configure_deltacat_logger(logging.getLogger(__name__))

__version__ = "0.2.7"


__all__ = [
    "__version__",
    "all_catalogs",
    "alter_table",
    "create_table",
    "drop_table",
    "refresh_table",
    "list_tables",
    "get_table",
    "truncate_table",
    "rename_table",
    "table_exists",
    "list_namespaces",
    "get_namespace",
    "namespace_exists",
    "alter_namespace",
    "create_namespace",
    "drop_namespace",
    "default_namespace",
    "write_to_table",
    "read_table",
    "init",
    "Catalog",
    "ContentType",
    "ContentEncoding",
    "DistributedDataset",
    "LifecycleState",
    "ListResult",
    "LocalDataset",
    "LocalTable",
    "Namespace",
    "SchemaConsistencyType",
    "SortKey",
    "SortOrder",
    "TableDefinition",
    "TableType",
    "TableWriteMode",
]



from typing import Any, Dict, List, Optional, Set, Union

import pyarrow as pa

from deltacat.catalog.model.table_definition import TableDefinition
from deltacat.storage.model.sort_key import SortKey
from deltacat.storage.model.list_result import ListResult
from deltacat.storage.model.namespace import Namespace
from deltacat.storage.model.types import (
    DistributedDataset,
    LifecycleState,
    LocalDataset,
    LocalTable,
    SchemaConsistencyType,
)
from deltacat.types.media import ContentType
from deltacat.types.tables import TableWriteMode


# table functions
def write_to_table(
    data: Union[LocalTable, LocalDataset, DistributedDataset],
    table: str,
    namespace: Optional[str] = None,
    mode: TableWriteMode = TableWriteMode.AUTO,
    content_type: ContentType = ContentType.PARQUET,
    *args,
    **kwargs
) -> None:
    """Write local or distributed data to a table. Raises an error if the
    table does not exist and the table write mode is not CREATE or AUTO.

    When creating a table, all `create_table` parameters may be optionally
    specified as additional keyword arguments. When appending to, or replacing,
    an existing table, all `alter_table` parameters may be optionally specified
    as additional keyword arguments."""
    raise NotImplementedError("write_to_table not implemented")


def read_table(
    table: str, namespace: Optional[str] = None, *args, **kwargs
) -> DistributedDataset:
    """Read a table into a distributed dataset."""
    raise NotImplementedError("read_table not implemented")


def alter_table(
    table: str,
    namespace: Optional[str] = None,
    lifecycle_state: Optional[LifecycleState] = None,
    schema_updates: Optional[Dict[str, Any]] = None,
    partition_updates: Optional[Dict[str, Any]] = None,
    primary_keys: Optional[Set[str]] = None,
    sort_keys: Optional[List[SortKey]] = None,
    description: Optional[str] = None,
    properties: Optional[Dict[str, str]] = None,
    *args,
    **kwargs
) -> None:
    """Alter table definition."""
    raise NotImplementedError("alter_table not implemented")


def create_table(
    table: str,
    namespace: Optional[str] = None,
    lifecycle_state: Optional[LifecycleState] = None,
    schema: Optional[Union[pa.Schema, str, bytes]] = None,
    schema_consistency: Optional[Dict[str, SchemaConsistencyType]] = None,
    partition_keys: Optional[List[Dict[str, Any]]] = None,
    primary_keys: Optional[Set[str]] = None,
    sort_keys: Optional[List[SortKey]] = None,
    description: Optional[str] = None,
    properties: Optional[Dict[str, str]] = None,
    permissions: Optional[Dict[str, Any]] = None,
    content_types: Optional[List[ContentType]] = None,
    replace_existing_table: bool = False,
    *args,
    **kwargs
) -> TableDefinition:
    """Create an empty table. Raises an error if the table already exists and
    `replace_existing_table` is False."""
    raise NotImplementedError("create_table not implemented")


def drop_table(
    table: str, namespace: Optional[str] = None, purge: bool = False, *args, **kwargs
) -> None:
    """Drop a table from the catalog and optionally purge it. Raises an error
    if the table does not exist."""
    raise NotImplementedError("drop_table not implemented")


def refresh_table(table: str, namespace: Optional[str] = None, *args, **kwargs) -> None:
    """Refresh metadata cached on the Ray cluster for the given table."""
    raise NotImplementedError("refresh_table not implemented")


def list_tables(
    namespace: Optional[str] = None, *args, **kwargs
) -> ListResult[TableDefinition]:
    """List a page of table definitions. Raises an error if the given namespace
    does not exist."""
    raise NotImplementedError("list_tables not implemented")


def get_table(
    table: str, namespace: Optional[str] = None, *args, **kwargs
) -> Optional[TableDefinition]:
    """Get table definition metadata. Returns None if the given table does not
    exist."""
    raise NotImplementedError("get_table not implemented")


def truncate_table(
    table: str, namespace: Optional[str] = None, *args, **kwargs
) -> None:
    """Truncate table data. Raises an error if the table does not exist."""
    raise NotImplementedError("truncate_table not implemented")


def rename_table(
    table: str, new_name: str, namespace: Optional[str] = None, *args, **kwargs
) -> None:
    """Rename a table."""
    raise NotImplementedError("rename_table not implemented")


def table_exists(table: str, namespace: Optional[str] = None, *args, **kwargs) -> bool:
    """Returns True if the given table exists, False if not."""
    raise NotImplementedError("table_exists not implemented")


# namespace functions
def list_namespaces(*args, **kwargs) -> ListResult[Namespace]:
    """List a page of table namespaces."""
    raise NotImplementedError("list_namespaces not implemented")


def get_namespace(namespace: str, *args, **kwargs) -> Optional[Namespace]:
    """Gets table namespace metadata for the specified table namespace. Returns
    None if the given namespace does not exist."""
    raise NotImplementedError("get_namespace not implemented")


def namespace_exists(namespace: str, *args, **kwargs) -> bool:
    """Returns True if the given table namespace exists, False if not."""
    raise NotImplementedError("namespace_exists not implemented")


def create_namespace(
    namespace: str, permissions: Dict[str, Any], *args, **kwargs
) -> Namespace:
    """Creates a table namespace with the given name and permissions. Returns
    the created namespace. Raises an error if the namespace already exists."""
    raise NotImplementedError("create_namespace not implemented")


def alter_namespace(
    namespace: str,
    permissions: Optional[Dict[str, Any]] = None,
    new_namespace: Optional[str] = None,
    *args,
    **kwargs
) -> None:
    """Alter table namespace definition."""
    raise NotImplementedError("alter_namespace not implemented")


def drop_namespace(namespace: str, purge: bool = False, *args, **kwargs) -> None:
    """Drop the given namespace and all of its tables from the catalog,
    optionally purging them."""
    raise NotImplementedError("drop_namespace not implemented")


def default_namespace() -> str:
    """Returns the default namespace for the catalog."""
    raise NotImplementedError("default_namespace not implemented")


# catalog functions
def initialize(*args, **kwargs) -> None:
    """Initializes the data catalog with the given arguments."""
    raise NotImplementedError("initialize not implemented")
